# ***************************************************************
# Copyright (c) 2019 Dun Liang <randonlang@gmail.com>. All Rights Reserved.
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import numpy as np
from .test_core import expect_error
from .test_grad import ngrad
from itertools import permutations
from jittor import compile_extern
from .test_log import find_log_with_re
if compile_extern.has_cuda:
    from jittor.compile_extern import cutt_ops
else:
    cutt_ops = None

def gen_data(shape):
    num = np.multiply.reduce(shape)
    a = np.arange(0, num)
    return a.reshape(shape)

class TestCuttTransposeOp(unittest.TestCase):
    @unittest.skipIf(cutt_ops==None, "Not use cutt, Skip")
    @jt.flag_scope(use_cuda=1)
    def test_with_np(self):
        def check(a):
            perms = list(permutations(range(a.ndim))) + [None]
            for perm in perms:
                with jt.log_capture_scope(
                    log_silent=1,
                    log_v=0, log_vprefix="op.cc=100"
                ) as raw_log:
                    if perm:
                        x = np.transpose(a, perm)
                        y = jt.transpose(a, perm).data
                    else:
                        x = np.transpose(a)
                        y = jt.transpose(a).data
                    self.assertEqual(x.shape, y.shape)
                logs = find_log_with_re(raw_log, "(Jit op key (not )?found: " + "cutt_transpose" + ".*)")
                if perm is None:
                    continue
                last = -1
                in_order = True
                for i in range(len(perm)):
                    if a.shape[perm[i]] == 1:
                        continue
                    if last != -1 and last > perm[i]:
                        in_order = False
                        break
                    last = perm[i]
                if not in_order:
                    assert len(logs)==1
                assert (x==y).all(), f"\n{x}\n{y}"
                
        ia = [gen_data([5, 7]), gen_data([2,2,2]), gen_data([2,3,4,5]), gen_data([5,3]), gen_data([3,1,5,3,1])]
        for a in ia: check(a)
        
    @unittest.skipIf(cutt_ops==None, "Not use cutt, Skip")
    @jt.flag_scope(use_cuda=1)
    def test_grad(self):
        def check(a):
            perms = list(permutations(range(a.ndim))) + [None]
            for perm in perms:
                x = jt.array(a).float()
                if perm:
                    y = jt.transpose(x, perm)
                else:
                    y = jt.transpose(x)
                dx = jt.grad(y*y, x).data
                self.assertEqual(dx.shape, a.shape)
                assert (dx==a*2).all(), f"\n{dx}\n{a}\n{perm}"
        ia = [gen_data([2,2,2]), gen_data([2,3,4,5]), gen_data([5,3]), gen_data([3,1,5,3,1])]
        for a in ia: check(a)
        
    @unittest.skipIf(cutt_ops==None, "Not use cutt, Skip")
    @jt.flag_scope(use_cuda=1)
    def test_matmul_grad(self):
        np.random.seed(0)
        for i in range(10):
            a = np.random.rand(2,3).astype("float32")
            b = np.random.rand(3,4).astype("float32")
            out, (da, db) = ngrad(lambda vars: np.matmul(vars[0],vars[1]).sum(), [a,b], 1e-1)
            ja = jt.array(a)
            jb = jt.array(b)
            jc = ja.matmul(jb)
            jda, jdb = jt.grad(jc, [ja,jb])
            assert ((da-jda.data)<1e-5).all(), (da, jda.data, da-jda.data)
            assert ((db-jdb.data)<1e-5).all(), (db-jdb.data)

if __name__ == "__main__":
    unittest.main()


# ***************************************************************
# Copyright (c) 2019 Dun Liang <randonlang@gmail.com>. All Rights Reserved.
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
from jittor.nn import Pool, pool
import numpy as np
from .test_core import expect_error
from .test_grad import ngrad
from itertools import permutations
from jittor import compile_extern, Module
from .test_log import find_log_with_re
import random
import pickle as pk

skip_this_test = False

try:
    jt.dirty_fix_pytorch_runtime_error()
    import torch
    from torch.nn import MaxPool2d, Sequential
except:
    skip_this_test = True

def check(jt_model, torch_model, shape, near_data):
    if (near_data):
        assert shape[0] * shape[1] * shape[2] * shape[3] % 8 == 0
        data = list(range(8)) * int((shape[0] * shape[1] * shape[2] * shape[3]) / 8)
        random.shuffle(data)
        x = jt.array(data).float32().reshape(shape)
    else:
        x = jt.random(shape)
    y = jt_model(x)
    g = jt.grad(y.sum(), x)

    x_ = torch.Tensor(x.data)
    x_.requires_grad = True
    y_ = torch_model(x_)
    y_.sum().backward()
    y__ = y_.detach().numpy()
    g__ = x_.grad.detach().numpy()
    assert np.allclose(y.data, y__)
    assert np.allclose(g.data, g__)

@unittest.skipIf(skip_this_test, "No Torch found")
class TestArgPoolOp(unittest.TestCase):
    @unittest.skipIf(not jt.compiler.has_cuda, "No cuda found")
    @jt.flag_scope(use_cuda=1)
    def test_cuda(self):
        jt_model = jt.nn.Sequential(Pool(2, 2, 0), Pool(2, 2, 0), Pool(2, 2, 0, ceil_mode=True), Pool(2, 2, 0), Pool(2, 2, 0), Pool(3, 1, 1))
        torch_model = Sequential(MaxPool2d(2, 2, 0), MaxPool2d(2, 2, 0), MaxPool2d(2, 2, 0, ceil_mode=True), MaxPool2d(2, 2, 0), MaxPool2d(2, 2, 0), MaxPool2d(3, 1, 1))
        shape = [64, 64, 300, 300]
        check(jt_model, torch_model, shape, False)
        shape = [32, 128, 157, 300]
        check(jt_model, torch_model, shape, False)
        for i in range(10):
            check(jt_model, torch_model, [1,1,300,300], True)
            
    def test_cpu_(self):
        # x = jt.random([32, 128, 157, 300])
        x = jt.random([4, 128, 157, 300])
        x = jt.nn.pool(x, 2, "maximum", 0, 2)

    def test_cpu(self):
        jt_model = jt.nn.Sequential(Pool(2, 2, 0), Pool(2, 2, 0), Pool(2, 2, 0, ceil_mode=True), Pool(2, 2, 0), Pool(2, 2, 0), Pool(3, 1, 1))
        torch_model = Sequential(MaxPool2d(2, 2, 0), MaxPool2d(2, 2, 0), MaxPool2d(2, 2, 0, ceil_mode=True), MaxPool2d(2, 2, 0), MaxPool2d(2, 2, 0), MaxPool2d(3, 1, 1))
        # shape = [64, 64, 300, 300]
        shape = [4, 64, 300, 300]
        check(jt_model, torch_model, shape, False)
        # shape = [32, 128, 157, 300]
        shape = [4, 128, 157, 300]
        check(jt_model, torch_model, shape, False)
        for i in range(10):
            check(jt_model, torch_model, [1,1,300,300], True)

if __name__ == "__main__":
    unittest.main()


# ***************************************************************
# Copyright (c) 2020 Jittor. Authors: Dun Liang <randonlang@gmail.com>. All Rights Reserved.
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import ast, astunparse

def convert(code):
    a = ast.parse(code)
    a.body.insert(0, ast.parse('import jittor as jt').body[0])
    a.body.insert(1, ast.parse('from jittor import init').body[0])
    dfs(a)
    return astunparse.unparse(a)

def replace(a):
    if hasattr(a, "attr"):
        if a.attr == "Conv2d": a.attr = "Conv"
        if a.attr == "BatchNorm2d": a.attr = "BatchNorm"
        if a.attr == "ReLU": a.attr = "Relu"
        if a.attr == "AvgPool2d": a.attr = "Pool"
        if a.attr == "MaxPool2d": a.attr = "Pool"
    if hasattr(a, "id"):
        if a.id == "Conv2d": a.id = "Conv"
        if a.id == "BatchNorm2d": a.id = "BatchNorm"
        if a.id == "ReLU": a.id = "Relu"
        if a.id == "AvgPool2d": a.id = "Pool"
        if a.id == "MaxPool2d": a.id = "Pool"

def dfs(a):
    if isinstance(a, ast.Import):
        if a.names[0].name == 'torch.nn' and a.names[0].asname == 'nn':
            a.names[0].name = 'jittor.nn'
            a.names[0].asname = 'nn'
    elif isinstance(a, ast.ImportFrom):
        if a.module == 'torch': 
            a.module = a.module.replace('torch', 'jittor')
            return a
    elif isinstance(a, ast.Call):
        for idx, ag in enumerate(a.args): 
            ret = dfs(ag)
            if ret is not None:
                a.args[idx] = ret
        for idx, kw in enumerate(a.keywords): 
            ret = dfs(kw)
            if ret is not None:
                a.keywords[idx] = ret
        if ".load_state_dict" in astunparse.unparse(a.func):
            a.func.attr = 'load_parameters'
        if astunparse.unparse(a.func).startswith("torch.Tensor"):
            a.func.value.id = 'jt'
            a.func.attr = 'array'
        if ".cat" in astunparse.unparse(a.func):
            if len(a.args) == 1:
                dim = a.keywords[0].value.n
            else:
                dim = a.args[1].n
            if isinstance(a.args[0], ast.List):
                objs = [elt.id for elt in a.args[0].elts]
                con = 'jt.contrib.concat([' + ','.join(objs) + '], dim=' + str(dim) + ')'
            else:
                con = 'jt.contrib.concat(' + a.args[0].id + ', dim=' + str(dim) + ')'
            return ast.parse(con).body[0].value
        if "view" in astunparse.unparse(a.func):
            ags = [astunparse.unparse(ag).strip('\n') for ag in a.args]
            con = 'jt.reshape(' + a.func.value.id + ', [' + ','.join(ags) + '])'
            return ast.parse(con).body[0].value
        if "permute" in astunparse.unparse(a.func):
            ags = [astunparse.unparse(ag).strip('\n') for ag in a.func.value.args]
            con = 'jt.transpose(' + a.func.value.func.value.id + ', [' + ','.join(ags) + '])'
            return ast.parse(con).body[0].value
        if astunparse.unparse(a.func).strip('\n').endswith(".size"):
            ags = [astunparse.unparse(ag).strip('\n') for ag in a.args]
            con = astunparse.unparse(a.func).split('.size')[0] + '.shape[' + ','.join(ags) + ']'
            return ast.parse(con).body[0].value
        if astunparse.unparse(a.func).startswith("F."):
            a.func.value.id = "nn"
            return a
        if "kaiming_normal_" in astunparse.unparse(a.func):
            ag = astunparse.unparse(a.args[0]).strip('\n')
            kws = {}
            for kw in a.keywords:
                tmp = astunparse.unparse(kw).split('=')
                kws[tmp[0]] = tmp[1].strip('\n')
            con = 'init.relu_invariant_gauss_(' + ag + ', mode=' + kws['mode'] + ')'
            return ast.parse(con).body[0].value
        if "constant_" in astunparse.unparse(a.func):
            ags = [astunparse.unparse(ag).strip('\n') for ag in a.args]
            con = 'init.constant_(' + ','.join(ags) + ')'
            return ast.parse(con).body[0].value
        if "ReLU" in astunparse.unparse(a.func):
            a.args.clear()
            a.keywords.clear()
        elif "Conv2d" in astunparse.unparse(a.func):
            pass
        elif "AvgPool2d" in astunparse.unparse(a.func):
            a.keywords.append(ast.keyword(arg='op', value=ast.Str(s='mean')))
        elif "MaxPool2d" in astunparse.unparse(a.func):
            a.keywords.append(ast.keyword(arg='op', value=ast.Str(s='maximum')))
        for kw in a.keywords:
            if kw.arg in ['return_indices', 'groups']:
                kw.value = ast.NameConstant(value=None)
    elif isinstance(a, ast.Expr): pass
    elif isinstance(a, ast.Attribute) or isinstance(a, ast.Name): replace(a)
    elif isinstance(a, ast.FunctionDef):
        if a.name == 'forward': a.name = 'execute'
    if hasattr(a, '__dict__'):
        for k in a.__dict__.keys():
            if isinstance(a.__dict__[k], list):
                for i,a_ in enumerate(a.__dict__[k]):
                    ret = dfs(a_)
                    if ret is not None:
                        a.__dict__[k][i] = ret
                            
            else:
                ret = dfs(a.__dict__[k])
                if ret is not None:
                    a.__dict__[k] = ret


# ***************************************************************
# Copyright (c) 2019 
#     Guoye Yang <498731903@qq.com>
#     Dun Liang <randonlang@gmail.com>. 
# All Rights Reserved.
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import numpy as np
from jittor import compile_extern
from .test_log import find_log_with_re
import copy
if compile_extern.has_cuda:
    from jittor.compile_extern import cutt_ops
else:
    cutt_ops = None

class TestCutt(unittest.TestCase):
    @unittest.skipIf(cutt_ops==None, "Not use cutt, Skip")
    @jt.flag_scope(use_cuda=1)
    def test(self):
        t = cutt_ops.cutt_test("213")
        assert t.data == 123
if __name__ == "__main__":
    unittest.main()


# ***************************************************************
# Copyright (c) 2020 Jittor. Authors:
#     Guowei Yang <471184555@qq.com>
#     Wenyang Zhou <576825820@qq.com>
#     Meng-Hao Guo <guomenghao1997@gmail.com>
#     Dun Liang <randonlang@gmail.com>.
#
# All Rights Reserved.
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import jittor as jt
from jittor import init, Module
import numpy as np
import math
from jittor.pool import Pool, pool

def matmul_transpose(a, b):
    '''
    returns a * b^T
    '''
    assert len(a.shape) >= 2 and len(b.shape) == 2
    assert a.shape[-1] == b.shape[-1]

    shape = list(a.shape)[:-1] + list(b.shape)
    a = a.broadcast(shape, [len(shape)-2])
    b = b.broadcast(shape)
    return (a*b).sum(len(shape)-1)

def matmul(a, b):
    assert len(a.shape) >= 2 and len(b.shape) == 2
    assert a.shape[-1] == b.shape[-2]

    shape = list(a.shape) + [b.shape[-1]]
    a = a.broadcast(shape, [len(shape)-1])
    b = b.broadcast(shape)
    return (a*b).sum(len(shape)-2)
jt.Var.matmul = jt.Var.__matmul__ = matmul
jt.Var.__imatmul__ = lambda a,b: a.assign(matmul(a,b))

def get_init_var_rand(shape, dtype):
    return jt.array(np.random.normal(0.0, 1.0, shape).astype(np.float32))

@jt.var_scope('batch_norm')
def batch_norm(x, is_train, eps=1e-5, momentum=0.1):
    w = jt.make_var([x.shape[1]], init=lambda *a: init.constant(*a, 1.0))
    b = jt.make_var([x.shape[1]], init=lambda *a: init.constant(*a, 0.0))
    running_mean = jt.make_var([x.shape[1]], init=lambda *a: init.constant(*a, 0.0))
    running_var = jt.make_var([x.shape[1]], init=lambda *a: init.constant(*a, 1.0))

    w = w.broadcast(x, [0,2,3])
    b = b.broadcast(x, [0,2,3])
    if is_train:
        xmean = jt.mean(x, dims=[0,2,3], keepdims=1)
        x2mean = jt.mean(x*x, dims=[0,2,3], keepdims=1)
        xvar = x2mean-xmean*xmean
        norm_x = (x-xmean)/jt.sqrt(xvar+eps)

        running_mean += (xmean.sum([0,2,3])-running_mean)*momentum
        running_var += (xvar.sum([0,2,3])-running_var)*momentum
    else:
        running_mean = running_mean.broadcast(x, [0,2,3])
        running_var = running_var.broadcast(x, [0,2,3])
        norm_x = (x-running_mean)/jt.sqrt(running_var+eps)

    return norm_x * w + b

@jt.var_scope('conv')
def conv(x, in_planes, out_planes, kernel_size, padding, stride = 1, init_method=None):
    Kw = kernel_size
    Kh = kernel_size
    _C = in_planes
    Kc = out_planes
    N,C,H,W = x.shape

    assert C==_C
    if init_method==None:
        w = jt.make_var([Kc, _C, Kh, Kw], init=lambda *a: init.relu_invariant_gauss(*a, mode="fan_out"))
    else:
        w = jt.make_var([Kc, _C, Kh, Kw], init=init_method)
    xx = x.reindex([N,Kc,C,(H+padding*2-kernel_size)//stride+1,(W+padding*2-kernel_size)//stride+1,Kh,Kw], [
        'i0', # Nid
        'i2', # Cid
        f'i3*{stride}-{padding}+i5', # Hid+Khid
        f'i4*{stride}-{padding}+i6', # Wid+KWid
    ])
    ww = w.broadcast(xx.shape, [0,3,4])
    yy = xx*ww
    y = yy.sum([2,5,6]) # C, Kh, Kw
    return y

@jt.var_scope('linear')
def linear(x, n):
    w = jt.make_var([n, x.shape[-1]], init=lambda *a: init.invariant_uniform(*a))
    w = w.reindex([w.shape[1], w.shape[0]],["i1","i0"])
    bound = 1.0/math.sqrt(w.shape[0])
    b = jt.make_var([n], init=lambda *a: init.uniform(*a,-bound,bound))
    return jt.matmul(x, w) + b

def relu(x): return jt.maximum(x, 0)
def leaky_relu(x, scale): return jt.ternary(x>0, x, x*scale)

#TODO dims is 4 will cause slowly execution
def cross_entropy_loss(output, target, ignore_index=None):
    if len(output.shape) == 4:
        c_dim = output.shape[1]
        output = output.transpose((0, 2, 3, 1))
        output = output.reshape((-1, c_dim))
    if ignore_index is not None:
        target = jt.ternary(target==ignore_index,
            jt.array(-1).broadcast(target), target)
        mask = jt.logical_and(target >= 0, target < output.shape[1])
    target = target.reshape((-1, ))
    target = target.broadcast(output, [1])
    target = target.index(1) == target

    output = output - output.max([1], keepdims=True)
    loss = output.exp().sum(1).log()
    loss = loss - (output*target).sum(1)
    if ignore_index is None:
        return loss.mean()
    else:
        return loss.sum() / jt.maximum(mask.int().sum(), 1)

class SGD(object):
    """ Usage:
    optimizer = nn.SGD(model.parameters(), lr)
    optimizer.step(loss)
    """
    def __init__(self, parameters, lr, momentum=0, weight_decay=0, dampening=0, nesterov=False):
        self.lr = lr
        self.momentum = momentum
        self.weight_decay = weight_decay
        self.dampening = dampening
        self.nesterov = nesterov

        self.no_grad_parameters = []
        self.parameters = []
        self.values = []
        for p in parameters:
            if p.is_stop_grad():
                self.no_grad_parameters.append(p)
                continue
            self.parameters.append(p)
            self.values.append(jt.zeros(p.shape, p.dtype).stop_fuse().stop_grad())

    def step(self, loss):
        ps = self.parameters
        gs = jt.grad(loss, ps)
        for p, g, v in zip(ps, gs, self.values):
            dp = p * self.weight_decay + g
            v.assign(self.momentum * v + dp * (1 - self.dampening))
            if self.nesterov:
                p -= (dp + self.momentum * v) * self.lr
            else:
                p -= v * self.lr
            # detach with the prev graph to reduce memory consumption
            p.detach_inplace()
        # sync all no grad parameters, such as
        # moving_mean and moving_var in batch_norm
        # sync such parameters to reduce memory consumption
        jt.sync(self.no_grad_parameters)

class Adam(object):
    """ Usage:
    optimizer = nn.Adam(model.parameters(), lr)
    optimizer.step(loss)
    """
    def __init__(self, parameters, lr, eps=1e-8, betas=(0.9, 0.999), weight_decay=0):
        self.lr = lr
        self.eps = eps
        self.betas = betas
        # self.weight_decay = weight_decay
        assert weight_decay==0, "weight_decay is not supported yet"
        self.adam_step = 0
        
        self.no_grad_parameters = []
        self.parameters = []
        self.values = []
        self.m = []
        for p in parameters:
            if p.is_stop_grad():
                self.no_grad_parameters.append(p)
                continue
            self.parameters.append(p)
            self.values.append(jt.zeros(p.shape, p.dtype).stop_fuse().stop_grad())
            self.m.append(jt.zeros(p.shape, p.dtype).stop_fuse().stop_grad())

    def step(self, loss):
        ps = self.parameters
        gs = jt.grad(loss, ps)
        self.adam_step += 1
        n, (b0, b1) = float(self.adam_step), self.betas
        for p, g, v, m in zip(ps, gs, self.values, self.m):
            m.assign(b0 * m + (1-b0) * g)
            v.assign(b1 * v + (1-b1) * g * g)
            step_size = self.lr * jt.sqrt(1-b1**n) / (1-b0 ** n)
            p -= m * step_size / (jt.sqrt(v) + self.eps)
            p.detach_inplace()
        jt.sync(self.no_grad_parameters)

def softmax(x, dim = None):
    if dim is None:
        x = (x - x.max()).exp()
        ret = x / x.sum()
    else:
        x = (x-x.max(dim, keepdims=True)).exp()
        ret = x / x.sum(dim, keepdims=True)
    return ret

class Dropout(Module):
    def __init__(self, p=0.5, is_train=False):
        assert p >= 0 and p <= 1, "dropout probability has to be between 0 and 1, but got {}".format(p)
        self.p = p
        self.is_train = is_train
        #TODO: test model.train() to change self.is_train
    def execute(self, input):
        output = input
        if self.p > 0 and self.is_train:
            if self.p == 1:
                noise = jt.zeros(input.shape)
            else:
                noise = jt.random(input.shape)
                noise = (noise > self.p).int()
            output = output * noise
        return output

class Linear(Module):
    def __init__(self, in_features, out_features, bias=True):
        self.in_features = in_features
        self.out_features = out_features
        self.weight = init.invariant_uniform((out_features, in_features), "float32")
        bound = 1.0/math.sqrt(in_features)
        self.bias = init.uniform((out_features,), "float32",-bound,bound) if bias else None

    def execute(self, x):
        x = matmul_transpose(x, self.weight)
        if self.bias is not None:
            return x + self.bias
        return x

class BatchNorm(Module):
    def __init__(self, num_features, eps=1e-5, momentum=0.1, affine=None, is_train=True):
        assert affine == None

        self.num_features = num_features
        self.is_train = is_train
        self.eps = eps
        self.momentum = momentum
        self.weight = init.constant((num_features,), "float32", 1.0)
        self.bias = init.constant((num_features,), "float32", 0.0)
        self.running_mean = init.constant((num_features,), "float32", 0.0).stop_grad()
        self.running_var = init.constant((num_features,), "float32", 1.0).stop_grad()

    def execute(self, x):
        if self.is_train:
            xmean = jt.mean(x, dims=[0,2,3], keepdims=1)
            x2mean = jt.mean(x*x, dims=[0,2,3], keepdims=1)
            xvar = x2mean-xmean*xmean
            norm_x = (x-xmean)/jt.sqrt(xvar+self.eps)
            self.running_mean += (xmean.sum([0,2,3])-self.running_mean)*self.momentum
            self.running_var += (xvar.sum([0,2,3])-self.running_var)*self.momentum
        else:
            running_mean = self.running_mean.broadcast(x, [0,2,3])
            running_var = self.running_var.broadcast(x, [0,2,3])
            norm_x = (x-running_mean)/jt.sqrt(running_var+self.eps)
        w = self.weight.broadcast(x, [0,2,3])
        b = self.bias.broadcast(x, [0,2,3])
        return norm_x * w + b

Relu = jt.make_module(relu)
ReLU = Relu
Leaky_relu = jt.make_module(leaky_relu, 2)
Softmax = jt.make_module(softmax, 2)

class Conv(Module):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1, padding=0, dilation=1, groups=1, bias=True):
        assert groups == 1

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size if isinstance(kernel_size, tuple) else (kernel_size, kernel_size)
        self.stride = stride if isinstance(stride, tuple) else (stride, stride)
        self.padding = padding if isinstance(padding, tuple) else (padding, padding)
        self.dilation = dilation if isinstance(dilation, tuple) else (dilation, dilation)
        Kh, Kw = self.kernel_size
        assert groups==1, "Group conv not supported yet."
        self.weight = init.relu_invariant_gauss([out_channels, in_channels, Kh, Kw], dtype="float", mode="fan_out")
        if bias:
            self.bias = init.uniform([out_channels], dtype="float", low=-1, high=1)
        else:
            self.bias = None

    def execute(self, x):
        N,C,H,W = x.shape
        Kh, Kw = self.kernel_size
        assert C==self.in_channels
        oh = (H+self.padding[0]*2-Kh*self.dilation[0]+self.dilation[0]-1)//self.stride[0]+1
        ow = (W+self.padding[1]*2-Kw*self.dilation[1]+self.dilation[1]-1)//self.stride[1]+1
        xx = x.reindex([N,self.out_channels,C,oh,ow,Kh,Kw], [
            'i0', # Nid
            'i2', # Cid
            f'i3*{self.stride[0]}-{self.padding[0]}+i5*{self.dilation[0]}', # Hid+Khid
            f'i4*{self.stride[1]}-{self.padding[1]}+i6*{self.dilation[1]}', # Wid+KWid
        ])
        ww = self.weight.broadcast(xx.shape, [0,3,4])
        yy = xx*ww
        y = yy.sum([2,5,6]) # Kc, Kh, Kw
        if self.bias is not None:
            b = self.bias.broadcast(y.shape, [0,2,3])
            y = y + b
        return y

class ConvTranspose(Module):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1, \
                 padding=0, output_padding=0, groups=1, bias=True, dilation=1):
        self.in_channels = in_channels
        self.out_channels = out_channels

        # added
        self.dilation = dilation
        self.group = groups
        assert groups==1, "Group conv not supported yet."

        self.kernel_size = kernel_size if isinstance(kernel_size, tuple) else (kernel_size, kernel_size)
        self.stride = stride if isinstance(stride, tuple) else (stride, stride)
        self.dilation = dilation if isinstance(dilation, tuple) else (dilation, dilation)
        # added
        self.padding = padding if isinstance(padding, tuple) else (padding, padding)
        self.real_padding = (self.dilation[0] * (self.kernel_size[0] - 1) - self.padding[0],
            self.dilation[1] * (self.kernel_size[1] - 1) - self.padding[1])
        self.output_padding = output_padding if isinstance (output_padding, tuple) else (output_padding, output_padding)

        self.weight = init.relu_invariant_gauss((in_channels, out_channels) + self.kernel_size, dtype="float", mode="fan_out")
        if bias:
            self.bias = init.uniform([out_channels], dtype="float", low=-1, high=1)
        else:
            self.bias = None

    def execute(self, x):
        N,C,H,W = x.shape
        i,o,h,w = self.weight.shape
        assert C==i
        stride_h, stride_w = self.stride
        padding_h, padding_w = self.padding
        dilation_h, dilation_w = self.dilation

        h_out = (H-1) * stride_h + self.output_padding[0] - 2*padding_h + 1 + (h-1)*dilation_h
        w_out = (W-1) * stride_w + self.output_padding[1] - 2*padding_w + 1 + (w-1)*dilation_w
        out_shape = (N, o, h_out, w_out)
        shape = (N, i, o, H, W, h, w)
        xx = x.broadcast(shape, (2, 5, 6)) # i,h,w
        ww = self.weight.broadcast(shape, (0, 3, 4)) # N,H,W
        y = (ww*xx).reindex_reduce("add", out_shape, [
            'i0', # N
            'i2', # o
            f'i3*{stride_h}-{padding_h}+i5*{dilation_h}', # Hid+Khid
            f'i4*{stride_w}-{padding_w}+i6*{dilation_w}', # Wid+KWid
        ])
        if self.bias is not None:
            b = self.bias.broadcast(y.shape, [0,2,3])
            y = y + b
        return y


class Tanh(Module):
    def __init__(self):
        super().__init__()
    def execute(self, x) :
        return x.tanh()

class Sigmoid(Module):
    def __init__(self):
        super().__init__()
    def execute(self, x) :
        return 1 / (1 + jt.exp(-x))

def resize(x, size, mode="nearest"):
    img = x
    n,c,h,w = x.shape
    H,W = size
    new_size = [n,c,H,W]
    nid, cid, hid, wid = jt.index(new_size)
    x = hid * ((h-1)/(H-1))
    y = wid * ((w-1)/(W-1))
    if mode=="nearest":
        return img.reindex([nid, cid, x.floor(), y.floor()])
    if mode=="bilinear":
        fx, fy = x.floor(), y.floor()
        cx, cy = fx+1, fy+1
        dx, dy = x-fx, y-fy
        a = img.reindex_var([nid, cid, fx, fy])
        b = img.reindex_var([nid, cid, cx, fy])
        c = img.reindex_var([nid, cid, fx, cy])
        d = img.reindex_var([nid, cid, cx, cy])
        dnx, dny = 1-dx, 1-dy
        ab = dx*b + dnx*a
        cd = dx*d + dnx*c
        o = ab*dny + cd*dy
        return o
    raise(f"Not support {interpolation}")



class Sequential(Module):
    def __init__(self, *args):
        self.layers = args
    def __getitem__(self, idx):
        return self.layers[idx]
    def execute(self, x):
        for layer in self.layers:
            x = layer(x)
        return x
    def dfs(self, parents, k, callback, callback_leave):
        n_children = len(self.layers)
        ret = callback(parents, k, self, n_children)
        if ret == False:
            return
        for k,v in enumerate(self.layers):
            parents.append(self)
            v.dfs(parents, k, callback, callback_leave)
            parents.pop()
        if callback_leave:
            callback_leave(parents, k, self, n_children)



# ***************************************************************
# Copyright (c) 2020 Jittor. Authors: 
#     Guoye Yang <498731903@qq.com>
#     Dun Liang <randonlang@gmail.com>. 
# All Rights Reserved.
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import numpy as np
from jittor import compile_extern
from .test_log import find_log_with_re
if compile_extern.has_cuda:
    from jittor.compile_extern import cublas_ops, cudnn_ops, cub_ops
else:
    cublas_ops = cudnn_ops = cub_ops = None

def check_argsort(shape, dim, descending = False):
    x = jt.random(shape)
    y, y_key = jt.argsort(x, dim=dim, descending=descending)
    v = []
    for i in range(len(shape)):
        if (i == dim):
            v.append(y)
        else:
            v.append(jt.index(shape, dim=i))
    yk = jt.reindex(x, v)
    yk_ = yk.data
    y_key_ = y_key.data
    x__ = x.data
    if descending:
        x__ = -x__
    yk__ = np.sort(x__, axis=dim)
    if descending:
        yk__ = -yk__
    assert np.allclose(y_key_, yk__)
    assert np.allclose(yk_, yk__)

def check_cub_argsort(shape, dim, descending = False):
    with jt.log_capture_scope(
        log_silent=1,
        log_v=0, log_vprefix="op.cc=100"
    ) as raw_log:
        x = jt.random(shape)
        y, y_key = jt.argsort(x, dim=dim, descending=descending)
        v = []
        for i in range(len(shape)):
            if (i == dim):
                v.append(y)
            else:
                v.append(jt.index(shape, dim=i))
        yk = jt.reindex(x, v)
        yk_ = yk.data
        y_key_ = y_key.data
    logs = find_log_with_re(raw_log, "(Jit op key (not )?found: " + "cub_argsort" + ".*)")
    assert len(logs)==1
    x__ = x.data
    if descending:
        x__ = -x__
    yk__ = np.sort(x__, axis=dim)
    if descending:
        yk__ = -yk__
    assert np.allclose(y_key_, yk__)
    assert np.allclose(yk_, yk__)

def check_backward(shape, dim, descending = False):
    x = jt.random(shape)
    y, y_key = jt.argsort(x, dim=dim, descending=descending)
    loss = (y_key * y_key).sum()
    gs = jt.grad(loss, x)
    assert np.allclose(x.data*2, gs.data)

class TestArgsortOp(unittest.TestCase):
    def test(self):
        check_argsort([5,5], 0, False)
        check_argsort([5,5], 0, True)
        check_argsort([5,5], 1, False)
        check_argsort([5,5], 1, True)
        check_argsort([12, 34, 56, 78], 1, True)
        check_argsort([12, 34, 56, 78], 3, True)
        check_argsort([12, 34, 56, 78], 2, False)
        check_argsort([12, 34, 56, 78], 0, False)

    def test_backward(self):
        check_backward([5,5], 0, False)
        check_backward([5,5], 0, True)
        check_backward([5,5], 1, False)
        check_backward([5,5], 1, True)
        check_backward([12, 34, 56, 78], 1, True)
        check_backward([12, 34, 56, 78], 3, True)
        check_backward([12, 34, 56, 78], 2, False)
        check_backward([12, 34, 56, 78], 0, False)

    def test_doc(self):
        assert "Argsort Operator" in jt.argsort.__doc__

    @unittest.skipIf(cub_ops==None, "Not use cub, Skip")
    @jt.flag_scope(use_cuda=1)
    def test_cub(self):
        check_cub_argsort([5,5], 0, False)
        check_cub_argsort([5,5], 0, True)
        check_cub_argsort([5,5], 1, False)
        check_cub_argsort([5,5], 1, True)
        check_cub_argsort([12, 34, 56, 78], 1, True)
        check_cub_argsort([12, 34, 56, 78], 3, True)
        check_cub_argsort([12, 34, 56, 78], 2, False)
        check_cub_argsort([12, 34, 56, 78], 0, False)

    @unittest.skipIf(cub_ops==None, "Not use cub, Skip")
    @jt.flag_scope(use_cuda=1)
    def test_cub_backward(self):
        check_backward([5,5], 0, False)
        check_backward([5,5], 0, True)
        check_backward([5,5], 1, False)
        check_backward([5,5], 1, True)
        check_backward([12, 34, 56, 78], 1, True)
        check_backward([12, 34, 56, 78], 3, True)
        check_backward([12, 34, 56, 78], 2, False)
        check_backward([12, 34, 56, 78], 0, False)

if __name__ == "__main__":
    unittest.main()


# ***************************************************************
# Copyright (c) 2020 Jittor. Authors:
#     Meng-Hao Guo <guomenghao1997@gmail.com>
#     Dun Liang <randonlang@gmail.com>.
# All Rights Reserved.
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************

import jittor as jt
import os
from six.moves import urllib
import hashlib
from tqdm import tqdm
import numpy as np
from collections.abc import Sequence, Mapping
from PIL import Image
from .. import lock

def ensure_dir(dir_path):
    if not os.path.isdir(dir_path):
        os.makedirs(dir_path)

def _progress():
    pbar = tqdm(total=None)

    def bar_update(block_num, block_size, total_size):
        """ reporthook
        @block_num: the num of downloaded data block
        @block_size: the size of data block
        @total_size: the total size of remote file
        """
        if pbar.total is None and total_size:
            pbar.total = total_size
        progress_bytes = block_num * block_size
        pbar.update(progress_bytes - pbar.n)

    return bar_update

@lock.lock_scope()
def download_url_to_local(url, filename, root_folder, md5):
    ensure_dir(root_folder)
    file_path = os.path.join(root_folder, filename)
    if check_file_exist(file_path, md5):
        print("Data file has been downloaded and verified")
    else:
        try:
            print('Downloading ' + url + ' to ' + file_path)
            urllib.request.urlretrieve(
                url, file_path,
                reporthook=_progress()
            )
        except(urllib.error.URLError, IOError) as e:
            raise e
    if not check_file_exist(file_path, md5):
        raise RuntimeError("File downloads failed.")



def check_file_exist(file_path, md5):
    if not os.path.isfile(file_path):
        return False
    if md5 is None:
        return True
    return check_md5(file_path, md5)


def calculate_md5(file_path, chunk_size=1024 * 1024):
    md5 = hashlib.md5()
    with open(file_path, 'rb') as f:
        for chunk in iter(lambda: f.read(chunk_size), b''):
            md5.update(chunk)
    return md5.hexdigest()


def check_md5(file_path, md5, **kwargs):
    return md5 == calculate_md5(file_path, **kwargs)


def get_random_list(n):
    return list(np.random.permutation(range(n)))

def get_order_list(n):
    return [i for i in range(n)]


def collate_batch(batch):
    r"""Puts each data field into a tensor with outer dimension batch size"""
    real_size = len(batch)
    elem = batch[0]
    elem_type = type(elem)
    if isinstance(elem, jt.Var):
        if elem.ndim == 1:
            temp_data = np.stack([data.numpy() for data in batch], 0)
            temp_data = np.squeeze(temp_data, -1)
            return jt.array(temp_data)
        else:
            temp_data = np.stack([data.numpy() for data in batch], 0)
        return jt.array(temp_data)
    if elem_type is np.ndarray:
        temp_data = np.stack([data for data in batch], 0)
        return temp_data
    elif np.issubdtype(elem_type, np.integer):
        return np.int32(batch)
    elif isinstance(elem, int):
        return np.int32(batch)
    elif isinstance(elem, float):
        return np.float32(batch)
    elif isinstance(elem, str):
        return batch
    elif isinstance(elem, Mapping):
        return {key: collate_batch([d[key] for d in batch]) for key in elem}
    elif isinstance(elem, tuple):
        transposed = zip(*batch)
        return tuple(collate_batch(samples) for samples in transposed)
    elif isinstance(elem, Sequence):
        transposed = zip(*batch)
        return [collate_batch(samples) for samples in transposed]
    elif isinstance(elem, Image.Image):
        temp_data = np.stack([np.array(data) for data in batch], 0)
        return temp_data
    else:
        raise TypeError(f"Not support type <{elem_type.__name__}>")






# ***************************************************************
# Copyright (c) 2020 Jittor. Authors: Dun Liang <randonlang@gmail.com>. All Rights Reserved.
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import numpy as np
from .test_core import expect_error

def equal_size(x, y):
    return x.dtype == y.dtype and x.shape == y.shape

def ngrad(func, vars, eps):
    out = func(vars)
    dout = []
    for i in range(len(vars)):
        pvar = vars[i].astype("float64")
        if type(pvar)==np.ndarray and pvar.size>1:
            grad = []
            var_f = pvar.flatten()
            for j in range(len(var_f)):
                var = pvar.flatten()
                var[j] += eps
                vars[i] = var.reshape(pvar.shape)
                out2 = func(vars)
                grad.append((out2-out)/eps)
            dout.append(np.array(grad).reshape(pvar.shape))
        else:
            vars[i] = vars[i] + eps
            out2 = func(vars)
            dout.append((out2-out)/eps)
        vars[i] = pvar
    return out, dout

class TestGrad(unittest.TestCase):
    def test_grad(self):
        x = jt.array([1.0, 2.0])
        y = jt.array([3.0, 4.0])
        z = x*y
        dx, dy, dz = jt.grad(z, [x,y,z])
        assert equal_size(dx, x) and equal_size(dy, y), f"{x} {y} {dx} {dy}"
        assert (dy.data == x.data).all(), f"{dy.data} {x.data}"
        assert (dx.data == y.data).all(), f"{dx.data} {y.data}"
        assert (dz.data == 1).all()
        
    def test_check_float(self):
        x = jt.array(1)
        y = x*x
        expect_error(lambda: jt.grad(y, [x]))
        
    def test_grad2(self):
        def test(n):
            x = jt.array(2.0)
            y = x
            for _ in range(n-1): y = y*x
            dx, = jt.grad(y, [x])
            assert dx.data == n*2**(n-1), f"{dx.data} {x.data}, {y.data}"
        test(5)
        test(6)
        test(7)
        test(8)
        
    def test_var_index(self):
        x = jt.array(2.0)
        y = x-x
        dx, = jt.grad(y, [x])
        assert dx.data == 0, dx.data
        x = jt.array(2.0)
        y = x/x
        dx, = jt.grad(x, [y])
        assert dx.data == 0
        
    def test_random_graph(self):
        def test(num_vars, num_ops, seed):
            np.random.seed(seed)
            vars = []
            for _ in range(num_vars):
                vars.append(np.random.rand(1))
            def random_func(vars):
                np.random.seed(seed+1)
                vars = list(vars)
                for i in range(num_ops):
                    v1 = len(vars)-1-np.random.randint(num_vars)
                    v2 = len(vars)-1-np.random.randint(num_vars)
                    rop = "+-*/"[np.random.randint(4)]
                    if (rop == '/' or rop == '-') and v1 is v2:
                        rop = '+'
                    vout = eval(f"vars[v1]{rop}vars[v2]")
                    vars.append(vout)
                if type(vars[i]) == jt.Var:
                    for i in range(len(vars)):
                        vars[i].name("v"+str(i))
                return vout
            np_out, np_dout = ngrad(random_func, vars, 1e-7)
            
            jt_vars = [ jt.array(v) for v in vars ]
            jt_out = random_func(jt_vars)
            assert (np.abs(jt_out.data-np_out) < 1e-5).all(), (jt_out.data, np_out)
            jt_dout = jt.grad(jt_out, jt_vars)
            jt_dout = [ v.data for v in jt_dout ]
            for jt_d, np_d in zip(jt_dout, np_dout):
                assert abs(jt_d - np_d) < 1e-3, f"{jt_d} {np_d}"
        test(1,1,0)
        # test(3,3,1)
        test(3,6,0)
        test(10,100,2)
        test(30,100,4)
        test(50,100,6)
        
    def test_top_sort(self):
        x = jt.array(2.0)
        x.name('x')
        y1 = x*x # 2
        y1.name('y1')
        y2 = x*x # 2
        y2.name('y2')
        y3 = y1*y2 # 4
        y3.name('y3')
        y4 = y3*y1 # 6
        y4.name('y4')
        y5 = y4*y1 # 8
        y5.name('y5')
        y6 = y5*y1 # 10
        y6.name('y6')
        vars = [x,y1,y2,y3,y4,y5,y6]
        grads = [ g.data for g in jt.grad(y6, vars) ]
        dx = grads[0]
        assert dx == 10*2**9, f"{grads}"
        
    def test_int_grad(self):
        x = jt.array(2.0)
        z = x*x*x*x*x
        dx, = jt.grad(z, [x])
        self.assertEqual(dx.data, 5*2**4)
        
        y1 = jt.int(x)
        y2 = jt.float(x)
        z = x*x*y1*y1*y2
        expect_error(lambda: jt.grad(z, [y1]))
        dx, = jt.grad(z, [x])
        self.assertEqual(dx.data, 48)
        
    def test_nth_grad(self):
        x = jt.array(2.0)
        y = x*x*x*x
        dx = jt.grad(y, x)
        ddx = jt.grad(dx, x)
        dddx = jt.grad(ddx, x)
        self.assertEqual(y.data, 2**4)
        self.assertEqual(dx.data, 4*2**3)
        self.assertEqual(ddx.data, 4*3*2**2)
        self.assertEqual(dddx.data, 4*3*2*2**1)
        

if __name__ == "__main__":
    unittest.main()



#!python3
import os, json
from pathlib import Path
notebook_dir = os.path.join(str(Path.home()), ".cache","jittor","notebook")
if not os.path.isdir(notebook_dir):
    os.mkdir(notebook_dir)
dirname = os.path.dirname(__file__)
all_md = []
for r, _, f in os.walk(dirname):
    for fname in f:
        if not fname.endswith(".md"): continue
        all_md.append(os.path.join(r, fname))
for mdname in all_md:
    with open(os.path.join(dirname, mdname), "r") as f:
        src = f.read()
    blocks = []
    for i, b in enumerate(src.split("```")):
        b = b.strip()
        is_markdown_block = i%2==0
        if not is_markdown_block and not b.startswith("python"):
            is_markdown_block = True
            b = "```\n"+b+"\n```"
        if is_markdown_block:
            # in a markdown block
            if len(blocks)%2==0:
                # prev code block
                blocks.append(b)
            else:
                # prev markdown block
                blocks[-1] += "\n\n" + b
        else:
            # in a code block
            if b.startswith("python"):
                b = b[6:].strip()
                # prev markdown block
                assert len(blocks)%2==1
                blocks.append(b)
    cells = []
    for i, b in enumerate(blocks):
        b = b.strip()
        if len(b)==0: continue
        b = b.split("\n")
        for j in range(len(b)-1):
            b[j] += '\n'
        cell = {
            "source": b,
            "metadata": {},
        }
        if i%2==0:
            cell["cell_type"] = "markdown"
        else:
            cell["cell_type"] = "code"
            cell["outputs"] = []
            cell["execution_count"] = None
        cells.append(cell)
    ipynb = {
        "cells":cells,
        "nbformat": 4,
        "nbformat_minor": 2,
        "metadata": {
        },
    }
    ipynb_name = mdname[:-2]+"ipynb"
    print(mdname, len(src), len(blocks), len(cells), "--->", ipynb_name)
    with open(os.path.join(notebook_dir, ipynb_name), "w") as f:
        f.write(json.dumps(ipynb))


# ***************************************************************
# Copyright (c) 2020 Jittor. Authors: Dun Liang <randonlang@gmail.com>. All Rights Reserved.
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import numpy as np

class TestCodeOp(unittest.TestCase):
    def test(self):
        a = jt.random([10])
        b = jt.code(a.shape, a.dtype, [a],
            cpu_src='''
                for (int i=0; i<in0shape0; i++)
                    @out(i) = @in0(i)*@in0(i)*2;
            ''',
            cpu_grad_src = ['''
                for (int i=0; i<in0shape0; i++) {
                    @out(i) = @dout(i)*@in0(i)*4;
                }
            '''])
        na, nb = jt.fetch_sync([a,b])
        assert np.allclose(na*na*2, nb)
        
        c = jt.random([10])
        da = jt.grad(c*b, a)
        assert np.allclose(c.data*na*4, da.data), (c.data*na*4, da.data)

    def test_multi_input(self):
        a = jt.random([10])
        b = jt.random([10])
        c = jt.code(a.shape, a.dtype, [a,b],
            cpu_src='''
                for (int i=0; i<in0shape0; i++)
                    @out(i) = @in0(i)*@in1(i);
            ''',
            cpu_grad_src = ['''
                for (int i=0; i<in0shape0; i++) {
                    @out(i) = @dout(i)*@in1(i);
                }
            ''', '''
                for (int i=0; i<in0shape0; i++) {
                    @out(i) = @dout(i)*@in0(i);
                }
            '''])
        da, db = jt.grad(c, [a, b])
        assert np.allclose(c.data, a.data*b.data), (c.data, a.data*b.data)
        assert np.allclose(da.data, b.data)
        assert np.allclose(db.data, a.data)

    def test_header(self):
        a = jt.array([3,2,1])
        b = jt.code(a.shape, a.dtype, [a],
            cpu_header='#include <algorithm>',
            cpu_src="""
                for (int i=0; i<in0shape0; i++)
                    @out(i) = @in0(i);
                std::sort(&@out(0), &@out(in0shape0));
            """
        )
        assert (b.data==[1,2,3]).all()

    @unittest.skipIf(not jt.compiler.has_cuda, "No CUDA found")
    @jt.flag_scope(use_cuda=1)
    def test_cuda(self):
        a = jt.random([100000])
        b = jt.random([100000])
        c = jt.code(a.shape, a.dtype, [a,b],
            cuda_src='''
            __global__ static void kernel1(@ARGS_DEF) {
                @PRECALC
                int i = threadIdx.x + blockIdx.x * blockDim.x;
                int stride = blockDim.x * gridDim.x;
                for (; i<in0shape0; i+=stride)
                    @out(i) = @in0(i)*@in1(i);
            }
                kernel1<<<(in0shape0-1)/1024+1, 1024>>>(@ARGS);
            ''',
            cuda_grad_src = ['''
            __global__ static void kernel2(@ARGS_DEF) {
                @PRECALC
                int i = threadIdx.x + blockIdx.x * blockDim.x;
                int stride = blockDim.x * gridDim.x;
                for (; i<in0shape0; i+=stride)
                    @out(i) = @dout(i)*@in1(i);
            }
                kernel2<<<(in0shape0-1)/1024+1, 1024>>>(@ARGS);
            ''', '''
            __global__ static void kernel3(@ARGS_DEF) {
                @PRECALC
                int i = threadIdx.x + blockIdx.x * blockDim.x;
                int stride = blockDim.x * gridDim.x;
                for (; i<in0shape0; i+=stride)
                    @out(i) = @dout(i)*@in0(i);
            }
                kernel3<<<(in0shape0-1)/1024+1, 1024>>>(@ARGS);
            '''])
        da, db = jt.grad(c, [a, b])
        assert np.allclose(c.data, a.data*b.data), (c.data, a.data*b.data)
        assert np.allclose(da.data, b.data)
        assert np.allclose(db.data, a.data)

    @unittest.skipIf(not jt.compiler.has_cuda, "No CUDA found")
    @jt.flag_scope(use_cuda=1)
    def test_cuda2(self):
        a = jt.random((100,100))
        b = jt.random((100,100))
        c = jt.code(a.shape, a.dtype, [a,b],
            cuda_src='''
                __global__ static void kernel1(@ARGS_DEF) {
                    @PRECALC
                    for (int i=blockIdx.x; i<in0shape0; i+=gridDim.x)
                    for (int j=threadIdx.x; j<in0shape1; j+=blockDim.x)
                        @out(i,j) = @in0(i,j)*@in1(i,j);
                }
                kernel1<<<32, 32>>>(@ARGS);
            ''',
            cuda_grad_src = ['''
                __global__ static void kernel(@ARGS_DEF) {
                    @PRECALC
                    for (int i=blockIdx.x; i<in0shape0; i+=gridDim.x)
                    for (int j=threadIdx.x; j<in0shape1; j+=blockDim.x)
                        @out(i,j) = @dout(i,j)*@in1(i,j);
                }
                kernel<<<32, 32>>>(@ARGS);
            ''', '''
                __global__ static void kernel(@ARGS_DEF) {
                    @PRECALC
                    @pout(0,0);
                    for (int i=blockIdx.x; i<in0shape0; i+=gridDim.x)
                    for (int j=threadIdx.x; j<in0shape1; j+=blockDim.x)
                        @out(i,j) = @dout(i,j)*@in0(i,j);
                }
                kernel<<<32, 32>>>(@ARGS);
            '''])
        da, db = jt.grad(c, [a, b])
        assert np.allclose(c.data, a.data*b.data), (c.data, a.data*b.data)
        assert np.allclose(da.data, b.data)
        assert np.allclose(db.data, a.data)


if __name__ == "__main__":
    unittest.main()


# ***************************************************************
# Copyright (c) 2020 Jittor. Authors: 
#     Guowei Yang <471184555@qq.com>
#     Dun Liang <randonlang@gmail.com>. 
# All Rights Reserved.
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import sys
import os
import jittor as jt
import unittest
import time
import numpy as np

def get_init_var(shape, dtype):
    return jt.random(shape, dtype)

def batch_norm(x):
    xmean = jt.mean(x, dims=[0,2,3], keepdims=1)
    x2mean = jt.mean(x*x, dims=[0,2,3], keepdims=1)
    norm_x = (x-xmean.broadcast_var(x))/(jt.sqrt(x2mean-xmean*xmean+jt.float32(1e-5)).broadcast_var(x))
    w = jt.make_var([x.shape[1]], init=get_init_var)
    b = jt.make_var([x.shape[1]], init=get_init_var)
    w = w.broadcast([1, w.shape[0],1,1], [0,2,3])
    b = b.broadcast([1, b.shape[0],1,1], [0,2,3])
    return norm_x * w + b

def pool(x, size, op, padding, stride = 1): # TODO: stride, padding
    N,C,H,W = x.shape
    h = (H+padding*2-size)//stride+1
    w = (W+padding*2-size)//stride+1
    xx = x.reindex([N,C,h,w,size,size], [
        "i0", # Nid
        "i1", # Cid
        f"i2*{stride}-{padding}+i4", # Hid
        f"i3*{stride}-{padding}+i5", # Wid
    ])
    return xx.reindex_reduce(op, [N,C,h,w], [
        "i0", # Nid
        "i1", # Cid
        "i2", # Hid
        "i3", # Wid
    ])

def conv(x, in_planes, out_planes, kernel_size, padding, stride = 1):
    Kw = kernel_size
    Kh = kernel_size
    _C = in_planes
    Kc = out_planes
    N,C,H,W = x.shape
    assert C==_C
    w = jt.make_var([Kc, _C, Kh, Kw], init=get_init_var)
    xx = x.reindex([N,Kc,C,(H+padding*2-kernel_size)//stride+1,(W+padding*2-kernel_size)//stride+1,Kh,Kw], [
        'i0', # Nid
        'i2', # Cid
        f'i3*{stride}-{padding}+i5', # Hid+Khid
        f'i4*{stride}-{padding}+i6', # Wid+KWid
    ])
    ww = w.broadcast(xx.shape, [0,3,4])
    yy = xx*ww
    y = yy.sum([2,5,6]) # Kc, Kh, Kw
    return y

def relu(x): return jt.maximum(x, jt.float32(0))

@jt.var_scope('resnet_fake', unique=True)
def resnet_fake(x):
    x = conv(x, 3, 64, 7, 3, 2)
    x = batch_norm(x)
    x = relu(x)
    x = pool(x, 3, "maximum", 1, 2)
    return x

class TestLongestDisFuse(unittest.TestCase):
        
    def test_longest_dis_fuse(self):
        x = jt.array(np.random.rand(1,3,224,224).astype(np.float32))
        loss = jt.sum(resnet_fake(x))
        ps = jt.find_vars('resnet_fake') 
        gs = jt.grad(loss, ps)
        jt.sync(gs)
        # assert not alloc big tensor
        g = jt.dump_all_graphs()
        for s in g.nodes_info:
            if not s.startswith("Var"):
                continue
            shape = s.split("[")[1].split("]")[0].split(",")
            ptr = s.split("(")[1].split(")")[0].split(",")[-1]
            if ptr != '0':
                assert len(shape)<=5, s

if __name__ == "__main__":
    unittest.main()



# ***************************************************************
# Copyright (c) 2020 Jittor. Authors: Dun Liang <randonlang@gmail.com>. All Rights Reserved.
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import math
import numpy as np

try:
    jt.dirty_fix_pytorch_runtime_error()
    import torch
    from torch import nn
    from jittor.utils import pytorch_converter
except:
    torch = None

@unittest.skipIf(torch is None, "pytorch not found.")
class TestPytorchConverter(unittest.TestCase):
    def test_simple(self):
        def model(c):
            a = torch.Tensor([1,2,3,4,0])
            b = a+a
            b = b*2
            b = b[:2]
            a = a[1<a]
            return a[0]+b[0]+c[0]

        c = torch.Tensor([1,2,3])
        r1 = model(c)
        with pytorch_converter.trace_scope(["model"]):
            r2 = model(c)
        assert r1.numpy()==r2.numpy()
        r3 = model(c)
        assert r1.numpy() == r2.numpy() and r2.numpy() == r3.numpy(), (r1,r2,r3)
        ans = """root in:[] out:[]
    model in:[input_0] out:[out_11]
        inj_torch_Tensor___init__ in:[array_1] out:[] args:[array_1, [1, 2, 3, 4, 0]]
        inj_torch_Tensor___add__ in:[array_1, array_1] out:[out_2] args:[array_1, array_1]
        inj_torch_Tensor___mul__ in:[out_2] out:[out_3] args:[out_2, 2]
        inj_torch_Tensor___getitem__ in:[out_3] out:[out_4] args:[out_3, slice(None, 2, None)]
        inj_torch_Tensor___gt__ in:[array_1] out:[out_5] args:[array_1, 1]
        inj_torch_Tensor___getitem__ in:[array_1, out_5] out:[out_6] args:[array_1, out_5]
        inj_torch_Tensor___getitem__ in:[out_6] out:[out_7] args:[out_6, 0]
        inj_torch_Tensor___getitem__ in:[out_4] out:[out_8] args:[out_4, 0]
        inj_torch_Tensor___add__ in:[out_7, out_8] out:[out_9] args:[out_7, out_8]
        inj_torch_Tensor___getitem__ in:[input_0] out:[out_10] args:[input_0, 0]
        inj_torch_Tensor___add__ in:[out_9, out_10] out:[out_11] args:[out_9, out_10]
    model in:[input_0] out:[out_11] end
root in:[] out:[] end"""
        ct = pytorch_converter.call_tree
        assert str(ct) == ans
        code = ct.to_jt()
        lc = {}
        exec(code, globals(), lc)
        print(code)
        jt_model = lc["model"]
        assert jt_model(jt.array([1,2,3])).data == r1.numpy()

    def test_resnet(self):
        class Bottleneck(nn.Module):
            expansion = 4

            def __init__(self, inplanes, planes, stride=1, downsample=None, dilation=1):
                super(Bottleneck, self).__init__()
                self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=False)
                self.bn1 = nn.BatchNorm2d(planes)
                self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride, dilation=dilation,
                                    padding=dilation, bias=False)
                self.bn2 = nn.BatchNorm2d(planes)
                self.conv3 = nn.Conv2d(planes, planes * 4, kernel_size=1, bias=False)
                self.bn3 = nn.BatchNorm2d(planes * 4)
                self.relu = nn.ReLU(inplace=True)
                self.downsample = downsample
                self.stride = stride

            def forward(self, x):
                residual = x

                out = self.conv1(x)
                out = self.bn1(out)
                out = self.relu(out)

                out = self.conv2(out)
                out = self.bn2(out)
                out = self.relu(out)

                out = self.conv3(out)
                out = self.bn3(out)

                if self.downsample is not None:
                    residual = self.downsample(x)

                out += residual
                out = self.relu(out)

                return out

        class ResNet(nn.Module):
            def __init__(self, block, layers=(3, 4, 23, 3)):
                self.inplanes = 64
                super(ResNet, self).__init__()
                self.conv1 = nn.Conv2d(3, 64, kernel_size=7, stride=2, padding=3,
                                    bias=False)
                self.bn1 = nn.BatchNorm2d(64)
                self.relu = nn.ReLU(inplace=True)
                self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
                self.layer1 = self._make_layer(block, 64, layers[0])
                self.layer2 = self._make_layer(block, 128, layers[1], stride=2)
                self.layer3 = self._make_layer(block, 256, layers[2], stride=1, dilation=2)
                self.layer4 = self._make_layer(block, 512, layers[3], stride=1, dilation=4)

                for m in self.modules():
                    if isinstance(m, nn.Conv2d):
                        n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                        m.weight.data.normal_(0, math.sqrt(2. / n))
                    elif isinstance(m, nn.BatchNorm2d):
                        m.weight.data.fill_(1)
                        m.bias.data.zero_()

            def _make_layer(self, block, planes, blocks, stride=1, dilation=1):
                downsample = None
                if stride != 1 or self.inplanes != planes * block.expansion:
                    downsample = nn.Sequential(
                        nn.Conv2d(self.inplanes, planes * block.expansion,
                                kernel_size=1, stride=stride, bias=False),
                        nn.BatchNorm2d(planes * block.expansion),
                    )

                layers = [block(self.inplanes, planes, stride, downsample)]
                self.inplanes = planes * block.expansion
                for i in range(1, blocks):
                    layers.append(block(self.inplanes, planes, dilation=dilation))

                return nn.Sequential(*layers)

            def forward(self, x):
                x1 = self.conv1(x)
                x2 = self.bn1(x1)
                x2 = self.relu(x2)
                x2 = self.maxpool(x2)

                x2 = self.layer1(x2)
                x3 = self.layer2(x2)
                x3 = self.layer3(x3)
                x3 = self.layer4(x3)

                return x1, x2, x3

        return
        resnet50 = ResNet(Bottleneck, [3, 4, 6, 3])
        x = torch.Tensor(np.random.rand(2, 3, 224, 224))
        with pytorch_converter.trace_scope():
            y = resnet50(x)

        ct = pytorch_converter.call_tree
        code = ct.to_jt()
        print(code)

    def test_convert_batchnorm(self):
        class TestModel(nn.Module):
            def __init__(self):
                super(TestModel, self).__init__()
                self.bn1 = nn.BatchNorm2d(64)
                self.bn2 = nn.BatchNorm2d(64)
                self.bn3 = nn.BatchNorm2d(64)
            def forward(self, x):
                y = self.bn1(x)
                z = self.bn2(x*x)
                x = self.bn3(y+z)
                return x
        model = TestModel()

        """ test_code:
            x = torch.Tensor(np.random.rand(16, 64, 15, 15).astype("float32"))
            jt_array = jt.array(x.numpy())
            jt_result = jt.nn.batch_norm(jt_array, is_train=False, eps=1e-05, momentum=0.1)
            torch_result = nn.BatchNorm2d(64)(x)
        """

        x = torch.Tensor(np.random.rand(16, 64, 15, 15).astype("float32"))
        with pytorch_converter.trace_scope():
            y = model(x)

        ct = pytorch_converter.call_tree
        ans = """root in:[] out:[]
    TestModel.forward in:[input_0] out:[out_27] args:{'self': TestModel(
  (bn1): BatchNorm2d(64, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)
  (bn2): BatchNorm2d(64, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)
  (bn3): BatchNorm2d(64, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)
), 'x': input_0}
        BatchNorm2d.forward in:[input_0] out:[out_7] args:{'self': BatchNorm2d(64, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True), 'input': input_0}
            functional.batch_norm in:[input_0, out_3, out_4, out_5, out_6] out:[out_7]
        BatchNorm2d.forward in:[input_0] out:[out_7] args:{'self': BatchNorm2d(64, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True), 'input': input_0} end
        inj_torch_Tensor___mul__ in:[input_0, input_0] out:[out_10] args:[input_0, input_0]
        BatchNorm2d.forward in:[out_10] out:[out_17] args:{'self': BatchNorm2d(64, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True), 'input': out_10}
            functional.batch_norm in:[out_10, out_13, out_14, out_15, out_16] out:[out_17]
        BatchNorm2d.forward in:[out_10] out:[out_17] args:{'self': BatchNorm2d(64, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True), 'input': out_10} end
        inj_torch_Tensor___add__ in:[out_7, out_17] out:[out_20] args:[out_7, out_17]
        BatchNorm2d.forward in:[out_20] out:[out_27] args:{'self': BatchNorm2d(64, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True), 'input': out_20}
            functional.batch_norm in:[out_20, out_23, out_24, out_25, out_26] out:[out_27]
        BatchNorm2d.forward in:[out_20] out:[out_27] args:{'self': BatchNorm2d(64, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True), 'input': out_20} end
    TestModel.forward in:[input_0] out:[out_27] args:{'self': TestModel(
  (bn1): BatchNorm2d(64, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)
  (bn2): BatchNorm2d(64, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)
  (bn3): BatchNorm2d(64, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)
), 'x': input_0} end
root in:[] out:[] end"""
        assert str(ct) == ans
        code = ct.to_jt()
        lc = {}
        exec(code, globals(), lc)
        print(code)
        
        jt_model = lc["TestModel"]
        assert (jt_model(jt.array(x.numpy())).data - y.detach().numpy()).mean() < 1e-5

    def test_convert_relu(self):
        class TestModel(nn.Module):
            def __init__(self):
                super(TestModel, self).__init__()
                self.rl1 = nn.ReLU(inplace=True)
                self.rl2 = nn.ReLU(inplace=True)
                self.rl3 = nn.ReLU(inplace=True)
            def forward(self, x):
                y = self.rl1(x)
                z = self.rl2(x*x)
                x = self.rl3(y+z)
                return x
        model = TestModel()

        """ test_code:
            x = torch.Tensor(np.random.rand(16, 64, 15, 15).astype("float32"))
            jt_array = jt.array(x.numpy())
            jt_result = jt.nn.batch_norm(jt_array, is_train=False, eps=1e-05, momentum=0.1)
            torch_result = nn.BatchNorm2d(64)(x)
        """

        x = torch.Tensor(np.random.rand(16, 3, 15, 15).astype("float32"))
        with pytorch_converter.trace_scope():
            y = model(x)

        ct = pytorch_converter.call_tree
        ans = """root in:[] out:[]
    TestModel.forward in:[input_0] out:[out_8] args:{'self': TestModel(
  (rl1): ReLU(inplace=True)
  (rl2): ReLU(inplace=True)
  (rl3): ReLU(inplace=True)
), 'x': input_0}
        ReLU.forward in:[input_0] out:[input_0] args:{'self': ReLU(inplace=True), 'input': input_0}
            functional.relu in:[input_0] out:[input_0]
        ReLU.forward in:[input_0] out:[input_0] args:{'self': ReLU(inplace=True), 'input': input_0} end
        inj_torch_Tensor___mul__ in:[input_0, input_0] out:[out_4] args:[input_0, input_0]
        ReLU.forward in:[out_4] out:[out_4] args:{'self': ReLU(inplace=True), 'input': out_4}
            functional.relu in:[out_4] out:[out_4]
        ReLU.forward in:[out_4] out:[out_4] args:{'self': ReLU(inplace=True), 'input': out_4} end
        inj_torch_Tensor___add__ in:[input_0, out_4] out:[out_8] args:[input_0, out_4]
        ReLU.forward in:[out_8] out:[out_8] args:{'self': ReLU(inplace=True), 'input': out_8}
            functional.relu in:[out_8] out:[out_8]
        ReLU.forward in:[out_8] out:[out_8] args:{'self': ReLU(inplace=True), 'input': out_8} end
    TestModel.forward in:[input_0] out:[out_8] args:{'self': TestModel(
  (rl1): ReLU(inplace=True)
  (rl2): ReLU(inplace=True)
  (rl3): ReLU(inplace=True)
), 'x': input_0} end
root in:[] out:[] end"""
        assert str(ct) == ans
        code = ct.to_jt()
        lc = {}
        exec(code, globals(), lc)
        print(code)
        
        jt_model = lc["TestModel"]
        assert (jt_model(jt.array(x.numpy())).data == y.detach().numpy()).all()

    def test_convert_pool(self):
        class TestModel(nn.Module):
            def __init__(self):
                super(TestModel, self).__init__()
                self.mp1 = nn.MaxPool2d(kernel_size=3, stride=2, padding=1, ceil_mode=False)
                self.mp2 = nn.MaxPool2d(kernel_size=3, stride=2, padding=1, ceil_mode=False)
                self.mp3 = nn.MaxPool2d(kernel_size=3, stride=2, padding=1, ceil_mode=False)
            def forward(self, x):
                y = self.mp1(x)
                z = self.mp2(x*x)
                x = self.mp3(y+z)
                return x
        model = TestModel()

        x = torch.Tensor(np.random.rand(16, 3, 15, 15).astype("float32"))
        with pytorch_converter.trace_scope():
            y = model(x)
        ct = pytorch_converter.call_tree
        ans = """root in:[] out:[]
    TestModel.forward in:[input_0] out:[out_11] args:{'self': TestModel(
  (mp1): MaxPool2d(kernel_size=3, stride=2, padding=1, dilation=1, ceil_mode=False)
  (mp2): MaxPool2d(kernel_size=3, stride=2, padding=1, dilation=1, ceil_mode=False)
  (mp3): MaxPool2d(kernel_size=3, stride=2, padding=1, dilation=1, ceil_mode=False)
), 'x': input_0}
        MaxPool2d.forward in:[input_0] out:[out_1] args:{'self': MaxPool2d(kernel_size=3, stride=2, padding=1, dilation=1, ceil_mode=False), 'input': input_0}
            functional._max_pool2d in:[input_0] out:[out_1]
        MaxPool2d.forward in:[input_0] out:[out_1] args:{'self': MaxPool2d(kernel_size=3, stride=2, padding=1, dilation=1, ceil_mode=False), 'input': input_0} end
        inj_torch_Tensor___mul__ in:[input_0, input_0] out:[out_5] args:[input_0, input_0]
        MaxPool2d.forward in:[out_5] out:[out_6] args:{'self': MaxPool2d(kernel_size=3, stride=2, padding=1, dilation=1, ceil_mode=False), 'input': out_5}
            functional._max_pool2d in:[out_5] out:[out_6]
        MaxPool2d.forward in:[out_5] out:[out_6] args:{'self': MaxPool2d(kernel_size=3, stride=2, padding=1, dilation=1, ceil_mode=False), 'input': out_5} end
        inj_torch_Tensor___add__ in:[out_1, out_6] out:[out_10] args:[out_1, out_6]
        MaxPool2d.forward in:[out_10] out:[out_11] args:{'self': MaxPool2d(kernel_size=3, stride=2, padding=1, dilation=1, ceil_mode=False), 'input': out_10}
            functional._max_pool2d in:[out_10] out:[out_11]
        MaxPool2d.forward in:[out_10] out:[out_11] args:{'self': MaxPool2d(kernel_size=3, stride=2, padding=1, dilation=1, ceil_mode=False), 'input': out_10} end
    TestModel.forward in:[input_0] out:[out_11] args:{'self': TestModel(
  (mp1): MaxPool2d(kernel_size=3, stride=2, padding=1, dilation=1, ceil_mode=False)
  (mp2): MaxPool2d(kernel_size=3, stride=2, padding=1, dilation=1, ceil_mode=False)
  (mp3): MaxPool2d(kernel_size=3, stride=2, padding=1, dilation=1, ceil_mode=False)
), 'x': input_0} end
root in:[] out:[] end"""
        assert str(ct) == ans
        code = ct.to_jt()
        lc = {}
        exec(code, globals(), lc)
        print(code)
        
        jt_model = lc["TestModel"]
        assert (jt_model(jt.array(x.numpy())).data == y.detach().numpy()).all()


if __name__ == "__main__":
    unittest.main()



import jittor as jt
import numpy as np
import sys, os
f32 = jt.float32

@jt.var_scope('linear')
def linear(x, n):
    w = jt.make_var([x.shape[-1], n], init=lambda *a:
            (jt.random(*a)-f32(0.5)) / f32(x.shape[-1])**f32(0.5))
    b = jt.make_var([n], init=lambda *a: jt.random(*a)-f32(0.5))
    return jt.matmul(x, w) + b
    
def relu(x): return jt.maximum(x, f32(0))

@jt.var_scope('model', unique=True)
def model(x):
    x = linear(x, 10)
    x = relu(x)
    x = linear(x, 1)
    return x

np.random.seed(0)
jt.set_seed(3)
n = 1000
batch_size = 50
base_lr = 0.05
# we need to stop grad of global value to prevent memory leak
lr = f32(base_lr).name("lr").stop_grad()

def get_data(n):
    for i in range(n):
        x = np.random.rand(batch_size, 1)
        y = x*x
        yield np.float32(x), np.float32(y)

for i,(x,y) in enumerate(get_data(n)):
    pred_y = model(x).name("pred_y")
    loss = ((pred_y - y)**f32(2)).name("loss")
    loss_mean = loss.mean()
    
    ps = jt.find_vars('model')
    gs = jt.grad(loss_mean, ps)
    for p,g in zip(ps, gs):
        p -= g * lr
    if i>2:
        assert prev == jt.liveness_info(), f"memory leak {prev} {jt.liveness_info()}"
    prev = jt.liveness_info()
    print(f"step {i}, loss = {loss_mean().sum()}")

# result is 0.0009948202641680837
result = 0.0009948202641680837
assert abs(loss_mean.data - result) < 1e-6


#!/usr/bin/python3
# ***************************************************************
# Copyright (c) 2020 Jittor. Authors: 
#     Guowei Yang <471184555@qq.com>
#     Dun Liang <randonlang@gmail.com>. 
# All Rights Reserved.
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import sys
import os
os.environ["log_silent"] = "1"
import re
import jittor_utils as jit_utils
from jittor_utils import LOG
jit_utils.try_import_jit_utils_core(silent=True)

def my_split(str):
    res=[]
    last=-1
    for i in range(len(str)):
        if str[i]==" " or str[i]=="\t":
            if last<i-1:
                res.append(str[last+1:i])
            last=i
        elif i==len(str)-1:
            res.append(str[last+1:])
    return res

def init(cc_path,s_path):
    global cc_content
    global s_content

    cc_content=[]
    s_content=[]
    with open(cc_path) as f:
        for line in f:
            cc_content.append(line)
    
    with open(s_path) as f:
        for line in f:
            s_content.append(line)

    global file_idx
    file_idx=0
    file_name=cc_path.split("/")[-1]

    idx=0
    for line in s_content:
        idx=idx+1
        if idx<=2:
            continue
        if ".file" in line and file_name in line:
            ss = my_split(line)
            if len(ss) == 2:
                file_idx = 0
                break
            file_idx=int(my_split(line)[1])
            break

def deal_replace(start,end,s1,s2):
    global s_content
    global file_idx

    for i in range(len(s_content)):
        line=s_content[i]
        if ".loc\t" in line or ".loc " in line:
            args=my_split(line)[1:]
            if int(args[0])==file_idx and start<=int(args[1]) and int(args[1])<=end:
                for j in range(i+1,len(s_content)):
                    if ".loc" in s_content[j]:
                        break
                    tmp=s_content[j]
                    s_content[j]=re.sub(s1,s2,s_content[j])
                    if tmp!=s_content[j]:
                        tmp=tmp.replace("\n","")
                        tmp=s_content[j]
                        tmp=tmp.replace("\n","")

def pass_asm(cc_path,s_path):
    global cc_content
    global s_content

    for i in range(len(cc_content)):
        line=cc_content[i]
        if "@begin" in line:
            cmds=line.split("@begin")[1].split(" ")
            si=0
            while cmds[si]=="":
                si=si+1
            
            start=i+1
            end=i+1
            for j in range(i+1,len(cc_content)):
                if "@end" in cc_content[j]:
                    end=j-1
                    break
            if cmds[si]=="replace":
                args=[]
                idx=0
                while line.find('"',idx)!=-1:
                    idx=line.find('"',idx)+1
                    args.append(line[idx:line.find('"',idx)])
                    idx=line.find('"',idx)+1
                deal_replace(start+1,end+1,args[0],args[1])
            else:
                assert 0, "no such command: "+line.split("@begin")[1]
    
    output_path=s_path.replace(".post.s",".s")
    with open(output_path,"w") as f:
        for line in s_content:
            f.write(line)

def run_cmd(cmd):
    LOG.vvvv(f"Run cmd: {cmd}")
    assert os.system(cmd) == 0, f"Run cmd failed: {cmd}"

args=sys.argv
args[1]=args[1][args[1].find("=")+1:]
compiler_path = sys.argv[1]
cc_path = sys.argv[2]

for i in range(len(args)):
    if args[i].find(" ")!=-1:
        args[i]="'"+args[i]+"'"
        
cmd = " ".join(args[1:])
cc_pos=cmd.find("_op.cc")
so_pos=cmd.find("_op.so")

# remove -Xclang ...
remove_clang_flag = lambda s: re.sub("-Xclang (('[^']*')|([^ ]*))", "", s)

def remove_flags(flags, rm_flags):
    flags = flags.split(" ")
    output = []
    for s in flags:
        if s.startswith("-load"):
            output.append(s)
            continue
        for rm in rm_flags:
            if s.startswith(rm):
                break
        else:
            output.append(s)
    return " ".join(output)

if cc_pos==-1:  #s_to_so
    run_cmd(remove_clang_flag(cmd))
elif so_pos==-1:  #cc_to_s
    asm_cmd=cmd.replace("_op.s", "_op.post.s") \
        .replace(" -o ", " -g -o ")
    run_cmd(asm_cmd)

    s_path = cc_path.replace("_op.cc","_op.post.s")
    init(cc_path,s_path)
    pass_asm(cc_path,s_path)
else:  #cc_to_so
    asm_cmd=cmd.replace("_op.so", "_op.post.s") \
        .replace("-lstdc++", "") \
        .replace("-ldl", "") \
        .replace("-shared", "-S") \
        .replace(" -o ", " -g -o ")
    asm_cmd = remove_flags(asm_cmd, ['-l', '-L', '-Wl,'])
    run_cmd(asm_cmd)

    s_path = cc_path.replace("_op.cc","_op.post.s")
    init(cc_path,s_path)
    pass_asm(cc_path,s_path)

    asm_cmd = cmd.replace("_op.cc", "_op.s") \
        .replace("-g", "")
    run_cmd(remove_clang_flag(asm_cmd))


# ***************************************************************
# Copyright (c) 2020 Jittor. Authors: 
#     Guowei Yang <471184555@qq.com>
#     Wenyang Zhou <576825820@qq.com>
#     Dun Liang <randonlang@gmail.com>. 
# All Rights Reserved.
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import jittor as jt
from jittor import nn
from jittor import Module

@jt.var_scope('basic_block')
def basic_block(x, is_train, in_planes, out_planes, stride = 1):
    identity = x
    x = nn.conv(x, in_planes, out_planes, 3, 1, stride)
    x = nn.batch_norm(x, is_train)
    x = nn.relu(x)
    x = nn.conv(x, out_planes, out_planes, 3, 1)
    x = nn.batch_norm(x, is_train)
    if in_planes!=out_planes:
        identity = nn.conv(identity, in_planes, out_planes, 1, 0, stride)
        identity = nn.batch_norm(identity, is_train)
    x = x+identity
    x = nn.relu(x)
    return x

@jt.var_scope('make_layer')
def make_layer(x, is_train, out_planes, blocks, layer_in_planes, stride = 1):
    x = basic_block(x, is_train, layer_in_planes, out_planes, stride)
    layer_in_planes = out_planes

    for i in range(1, blocks):
        x = basic_block(x, is_train, layer_in_planes, out_planes)
    return x, layer_in_planes

@jt.var_scope('bottleneck_block')
def bottleneck_block(x, is_train, in_planes, out_planes, stride = 1):
    expansion = 4 
    width = out_planes
    identity = x

    x = nn.conv(x, in_planes, width, 1, 0)
    x = nn.batch_norm(x, is_train)
    x = nn.relu(x)

    x = nn.conv(x, width, width, 3, 1, stride)
    x = nn.batch_norm(x, is_train)
    x = nn.relu(x)

    x = nn.conv(x, width, out_planes * expansion, 1, 0)
    x = nn.batch_norm(x, is_train)

    if in_planes != out_planes * expansion:
        identity = nn.conv(identity, in_planes, out_planes * expansion, 1, 0, stride)
        identity = nn.batch_norm(identity, is_train)
    
    x = x+identity
    x = nn.relu(x)
    return x

@jt.var_scope('make_layer_bottleneck')
def make_layer_bottleneck(x, is_train, out_planes, blocks, layer_in_planes, stride = 1):
    expansion = 4
    x = bottleneck_block(x, is_train, layer_in_planes, out_planes, stride)
    layer_in_planes = out_planes * expansion
    for i in range(1, blocks):
        x = bottleneck_block(x, is_train, layer_in_planes, out_planes)
    return x, layer_in_planes

@jt.var_scope('resnet')
def resnet(x, is_train, block, layers, num_classes = 1000):
    layer_in_planes = 64
    x = nn.conv(x, 3, layer_in_planes, 7, 3, 2)
    x = nn.batch_norm(x, is_train)
    x = nn.relu(x)
    x = nn.pool(x, 3, "maximum", 1, 2)
    x, layer_in_planes = block(x, is_train, 64, layers[0], layer_in_planes)
    x, layer_in_planes = block(x, is_train, 128, layers[1], layer_in_planes, 2)
    x, layer_in_planes = block(x, is_train, 256, layers[2], layer_in_planes, 2)
    x, layer_in_planes = block(x, is_train, 512, layers[3], layer_in_planes, 2)

    x = x.reindex_reduce("add", [x.shape[0],x.shape[1]], ["i0","i1"])/x.shape[2]/x.shape[3]
    x = nn.linear(x, num_classes)

    return x

@jt.var_scope('resnet18', unique=True)
def resnet18(x, is_train):
    return resnet(x, is_train, make_layer, [2, 2, 2, 2])

@jt.var_scope('resnet34', unique=True)
def resnet34(x, is_train):
    return resnet(x, is_train, make_layer, [3, 4, 6, 3])

@jt.var_scope('resnet50', unique=True)
def resnet50(x, is_train):
    return resnet(x, is_train, make_layer_bottleneck, [3, 4, 6, 3])

@jt.var_scope('resnet101', unique=True)
def resnet101(x, is_train):
    return resnet(x, is_train, make_layer_bottleneck, [3, 4, 23, 3])

@jt.var_scope('resnet152', unique=True)
def resnet152(x, is_train):
    return resnet(x, is_train, make_layer_bottleneck, [3, 8, 36, 3])

class BasicBlock(Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        self.conv1 = nn.Conv(inplanes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn1 = nn.BatchNorm(planes)
        self.relu = nn.Relu()
        self.conv2 = nn.Conv(planes, planes, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn2 = nn.BatchNorm(planes)
        self.downsample = downsample
        self.stride = stride
        self.planes = planes

    def execute(self, x):
        residual = x
        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)
        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            residual = self.downsample(x)
        
        out += residual
        out = self.relu(out)
        return out

class Bottleneck(Module):
    expansion = 4
 
    def __init__(self, inplanes, planes, stride=1, downsample=None):
        self.conv1 = nn.Conv(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm(planes)
        self.conv2 = nn.Conv(planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn2 = nn.BatchNorm(planes)
        self.conv3 = nn.Conv(planes, planes * self.expansion, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm(planes * self.expansion)
        self.relu = nn.Relu()
        self.downsample = downsample
        self.stride = stride
 
    def execute(self, x):
        residual = x
 
        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)
 
        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)
 
        out = self.conv3(out)
        out = self.bn3(out)
 
        if self.downsample is not None:
            residual = self.downsample(x)
 
        out += residual
        out = self.relu(out)
        return out

class ResNet(Module):
    def __init__(self, block, layers, num_classes=1000):
        self.inplanes = 64
        self.conv1 = nn.Conv(3, 64, kernel_size=7, stride=2, padding=3, bias=False)
        self.bn1 = nn.BatchNorm(64)
        self.relu = nn.Relu()
        self.maxpool = nn.Pool(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64, layers[0])
        self.layer2 = self._make_layer(block, 128, layers[1], stride=2)
        self.layer3 = self._make_layer(block, 256, layers[2], stride=2)
        self.layer4 = self._make_layer(block, 512, layers[3], stride=2)
        self.avgpool = nn.Pool(7, stride=1, op="mean")
        self.fc = nn.Linear(512 * block.expansion, num_classes)
 
    def _make_layer(self, block, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv(self.inplanes, planes * block.expansion,
                          kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm(planes * block.expansion),
            )
 
        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes))
 
        return nn.Sequential(*layers)
 
    def execute(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = jt.reshape(x, [x.shape[0],-1])
        x = self.fc(x)
 
        return x

def Resnet18():
    model = ResNet(BasicBlock, [2,2,2,2])
    return model

def Resnet34():
    model = ResNet(BasicBlock, [3,4,6,3])
    return model

def Resnet50():
    model = ResNet(Bottleneck, [3,4,6,3])
    return model

def Resnet101():
    model = ResNet(Bottleneck, [3,4,23,3])
    return model

def Resnet152():
    model = ResNet(Bottleneck, [3,8,36,3])
    return model


# ***************************************************************
# Copyright (c) 2020
#     Dun Liang <randonlang@gmail.com>. 
# All Rights Reserved.
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
from PIL import Image
import random
import math
import numpy as np
import warnings
from collections.abc import Sequence, Mapping

def crop(img, top, left, height, width):
    return img.crop((left, top, left + width, top + height))

def resize(img, size, interpolation=Image.BILINEAR):
    return img.resize(size[::-1], interpolation)

def crop_and_resize(img, top, left, height, width, size, interpolation=Image.BILINEAR):
    img = crop(img, top, left, height, width)
    img = resize(img, size, interpolation)
    return img

class RandomCropAndResize:
    """Random crop and resize the given PIL Image to given size.

    Args:
        size(int or tuple): width and height of the output image
        scale(tuple): range of scale ratio of the area
        ratio(tuple): range of aspect ratio
        interpolation: Default: PIL.Image.BILINEAR
    """
    def __init__(self, size, scale:tuple=(0.08, 1.0), ratio:tuple=(3. / 4., 4. / 3.), interpolation=Image.BILINEAR):
        if isinstance(size, int):
            size = (size, size)
        assert isinstance(size, tuple)
        assert scale[0] <= scale[1] and ratio[0] <= ratio[1]

        self.size = size
        self.scale = scale
        self.ratio = ratio
        self.interpolation = interpolation

    def __call__(self, img:Image.Image):
        width, height = img.size
        scale = self.scale
        ratio = self.ratio
        area = height * width

        for _ in range(10):
            target_area = random.uniform(*scale) * area
            log_ratio = (math.log(ratio[0]), math.log(ratio[1]))
            aspect_ratio = math.exp(random.uniform(*log_ratio))

            w = int(round(math.sqrt(target_area * aspect_ratio)))
            h = int(round(math.sqrt(target_area / aspect_ratio)))

            if 0 < w <= width and 0 < h <= height:
                i = random.randint(0, height - h)
                j = random.randint(0, width - w)
                break
        else:
            # Fallback to central crop
            in_ratio = float(width) / float(height)
            if in_ratio < min(ratio):
                w = width
                h = int(round(w / min(ratio)))
            elif in_ratio > max(ratio):
                h = height
                w = int(round(h * max(ratio)))
            else:
                w = width
                h = height
            i = (height - h) // 2
            j = (width - w) // 2
        return crop_and_resize(img, i, j, h, w, self.size, self.interpolation)
    
class RandomHorizontalFlip:
    def __init__(self, p=0.5):
        self.p = p
        
    def __call__(self, img:Image.Image):
        if random.random() < self.p:
            return img.transpose(Image.FLIP_LEFT_RIGHT)
        return img
    
def to_tensor(img):
    if isinstance(img, Image.Image):
        return np.array(img).transpose((2,0,1)) / np.float32(255)
    return img

class ImageNormalize:
    def __init__(self, mean, std):
        self.mean = np.float32(mean).reshape(-1,1,1)
        self.std = np.float32(std).reshape(-1,1,1)
        
    def __call__(self, img):
        if isinstance(img, Image.Image):
            img = (np.array(img).transpose((2,0,1)) \
                - self.mean*np.float32(255.)) \
                / (self.std*np.float32(255.))
        else:
            img = (img - self.mean) / self.std
        return img
    
class Compose:
    def __init__(self, transforms):
        self.transforms = transforms
    def __call__(self, data):
        for t in self.transforms:
            data = t(data)
        return data

class Resize:
    def __init__(self, size):
        if isinstance(size, int):
            size = (size, size)
        assert isinstance(size, tuple)
        self.size = size
    def __call__(self, img:Image.Image):
        return img.resize(self.size, Image.BILINEAR)

class Gray:
    def __call__(self, img:Image.Image):
        img = np.array(img.convert('L'))
        img = img[np.newaxis, :]
        return np.array((img / 255.0), dtype = np.float32)



# ***************************************************************
# Copyright (c) 2020 Jittor. Authors: 
#     Guoye Yang <498731903@qq.com>
#     Dun Liang <randonlang@gmail.com>. 
# All Rights Reserved.
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import numpy as np
from .test_grad import ngrad
from .test_cuda import test_cuda

def get_node_info(s):
    mem_ptr = s.split(')')[0].split(',')[-1]
    name = s.split(')')[0].split(',')[-2]
    return name, mem_ptr

def get_info(graph):
    bop = [ node for node in graph.nodes_info if node.startswith("Var")]
    node_dict = {}
    for bop_ in bop:
        name, mem_ptr = get_node_info(bop_)
        node_dict[name] = mem_ptr
    return node_dict

def check_equal(a, b):
    eps = 1e-1 # icc error almost reaches 1e-1
    return abs(a - b) < eps

class TestReshapeOp(unittest.TestCase):
    def test_reshape(self):
        a = jt.random([123, 456, 789]).name("a")
        b = jt.reshape(a, [123 * 2, int(789 * 456 / 2)]).name("b")
        c = jt.reshape(b, [123 * 456 * 789]).name("c")
        d = jt.reshape(c, [2, int(123 / 3), 789, int(456 / 2), 3]).name("d")
        e = jt.reshape(d, [2, int(123 / 3), 789, -1, 3]).name("e")
        assert b.shape == [123 * 2, int(789 * 456 / 2)]
        assert c.shape == [123 * 456 * 789]
        assert d.shape == [2, int(123 / 3), 789, int(456 / 2), 3]
        assert e.shape == [2, int(123 / 3), 789, int(456 / 2), 3]
        a_mean = a.mean().data
        b_mean = b.mean().data
        c_mean = c.mean().data
        d_mean = d.mean().data
        e_mean = e.mean().data
        a = (a + 1).name("new_a")
        new_a_mean = a.mean().data
        new_b_mean = b.mean().data
        node_dict = get_info(jt.dump_all_graphs())
        assert check_equal(a_mean, b_mean), f"{a_mean} != {b_mean}"
        assert check_equal(a_mean, c_mean), f"{a_mean} != {c_mean}"
        assert check_equal(a_mean, d_mean), f"{a_mean} != {d_mean}"
        assert check_equal(a_mean, e_mean), f"{a_mean} != {e_mean}"
        assert check_equal(b_mean, new_b_mean), f"{b_mean} != {new_b_mean}"
        assert not check_equal(a_mean, new_a_mean), f"{a_mean} == {new_a_mean}"
        assert node_dict['a'] == node_dict['b']
        assert node_dict['a'] == node_dict['c']
        assert node_dict['a'] == node_dict['d']
        assert node_dict['a'] == node_dict['e']

if __name__ == "__main__":
    unittest.main()


# ***************************************************************
# Copyright (c) 2020 Jittor. Authors: Dun Liang <randonlang@gmail.com>. All Rights Reserved.
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import os
from jittor import compile_extern
if compile_extern.has_cuda:
    from jittor.compile_extern import cublas_ops, cudnn_ops, cub_ops
else:
    cublas_ops = cudnn_ops = cub_ops = None

@unittest.skipIf(cublas_ops==None, "Not use cublas, Skip")
class TestCublasTestOp(unittest.TestCase):
    def test(self):
        assert cublas_ops.cublas_test(2).data==123
        assert cublas_ops.cublas_test(5).data==123
        assert cublas_ops.cublas_test(10).data==123
        assert cublas_ops.cublas_test(20).data==123

@unittest.skipIf(cudnn_ops==None, "Not use cudnn, Skip")
class TestCudnnTestOp(unittest.TestCase):
    def test(self):
        assert cudnn_ops.cudnn_test("").data == 123
        assert cudnn_ops.cudnn_test("-c2048 -h7 -w7 -k512 -r1 -s1 -pad_h0 -pad_w0 -u1 -v1").data == 123
        
@unittest.skipIf(cub_ops==None, "Not use cub, Skip")
class TestCubTestOp(unittest.TestCase):
    @jt.flag_scope(use_cuda=1)
    def test(self):
        assert cub_ops.cub_test("xx").data == 123
        assert cub_ops.cub_test("xx --n=100000").data == 123
        
if __name__ == "__main__":
    unittest.main()



# ***************************************************************
# Copyright (c) 2020 Jittor. Authors: Dun Liang <randonlang@gmail.com>. All Rights Reserved.
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import numpy as np
from .test_core import expect_error
ops = jt.ops

@jt.var_scope('linear')
def linear(x, n):
    w = jt.make_var([x.shape[-1], n], init=ops.random)
    return jt.matmul(x, w)

@jt.var_scope('model', unique=True)
def model(x):
    x = linear(x, 10)
    # x = relu(x)
    x = linear(x, 10)
    # x = relu(x)
    x = linear(x, 1)
    return x

class TestScope(unittest.TestCase):
    def test_name(self):
        jt.clean()
        @jt.var_scope('model', unique=True)
        def model():
            with jt.var_scope('a'):
                assert jt.current_scope.full_name == "model/a_0/"
            with jt.var_scope('b'):
                with jt.var_scope('b'):
                    assert jt.current_scope.full_name == "model/b_0/b_0/"
            with jt.var_scope('c'):
                assert jt.current_scope.full_name == "model/c_0/"
        model()
        model()
        model()
        jt.clean()
        
    def test_var(self):
        jt.clean()
        for i in range(2):
            x = jt.array([[1]])
            y = model(x)
            params = jt.find_vars()
            assert len(params) == 3
            names = [ p.name() for p in params ]
            assert names == [
                "model/linear_0/var_0",
                "model/linear_1/var_0",
                "model/linear_2/var_0",
            ], str(names)
            jt.find_var("model/linear_0/var_0")
            expect_error(lambda: jt.find_var("model/linear_0"))
            expect_error(lambda: jt.find_var("model/linear"))
            assert len(jt.find_vars("model/linear_0/var_0")) == 1
            assert len(jt.find_vars("model/linear_0/")) == 1
            assert len(jt.find_vars("model/")) == 3
        jt.clean()
        
    def test_get_var_unique(self):
        jt.clean()
        x = jt.make_var([1], init=ops.random)
        y = jt.make_var([1], init=ops.random)
        z = jt.make_var([1], init=ops.random)
        assert x.name() == "var_0"
        assert y.name() == "var_1", y.name()
        assert z.name() == "var_2"
        x = jt.make_var([1], name="x", unique=True, init=ops.random)
        y = jt.make_var([1], name="y", unique=True, init=ops.random)
        z = jt.make_var([1], name="z", unique=True, init=ops.random)
        assert x.name() == "x"
        assert y.name() == "y"
        assert z.name() == "z"
        expect_error(lambda: jt.make_var([2], name="x", unique=True, init=ops.random))
        jt.clean()

    def test_record_scope(self):
        jt.clean()
        @jt.var_scope("func")
        def func(a):
            b = a+1
            jt.record_in_scope(b, "b")
            c = b*2
            return c
        a = jt.array([1,2,3])
        func(a)
        assert np.allclose(jt.find_record("func_0/output").data, (a.data+1)*2)
        assert np.allclose(jt.find_scope("func_0").records["output"].data, (a.data+1)*2)
        recs = jt.find_records()
        rec_names = [ r.name() for r in recs ]
        assert len(recs)==2 and rec_names==["func_0/b","func_0/output"]
        
    def test_get_var_init(self):
        jt.clean()
        assert (jt.make_var(init=[1,2,3]).data == [1,2,3]).all()
        assert (jt.make_var(shape=[3], init=np.zeros).data == [0,0,0]).all()
        assert (jt.make_var(init=jt.array([1,2,3]) == [1,2,3]).data).all()
        jt.clean()

if __name__ == "__main__":
    unittest.main()


# ***************************************************************
# Copyright (c) 2020 Jittor. Authors: 
#     Guowei Yang <471184555@qq.com>
#     Guoye Yang <498731903@qq.com>
#     Dun Liang <randonlang@gmail.com>. 
# All Rights Reserved.
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import jittor as jt
import numpy as np

def argmax_pool(x, size, stride, padding=0):
    y_shape = list(x.shape)
    y_shape[2]=(x.shape[2]+padding*2-size)//stride+1
    y_shape[3]=(x.shape[3]+padding*2-size)//stride+1
    
    y = jt.code(y_shape, x.dtype, [x],
        cpu_src=f'''
            for (int i=0; i<outshape0; i++)
            for (int j=0; j<outshape1; j++)
            for (int k=0; k<outshape2; k++)
            for (int l=0; l<outshape3; l++) {{
                int kx=k*{stride}+{size}/2-{padding};
                int ky=l*{stride}+{size}/2-{padding};
                @out(i,j,k,l) = @in0(i,j,kx,ky);
                for (int p=kx-{size}/2;p<=kx+{size}/2;p++)
                for (int q=ky-{size}/2;q<=ky+{size}/2;q++)
                    if (p>=0 && q>=0 && p<in0shape2 && q<in0shape3)
                    if (@out(i,j,k,l) < @in0(i,j,p,q))
                        @out(i,j,k,l) = @in0(i,j,p,q);
            }}
        ''',
        cpu_grad_src = [f'''
            for (int i=0; i<outshape0; i++)
            for (int j=0; j<outshape1; j++)
            for (int k=0; k<outshape2; k++)
            for (int l=0; l<outshape3; l++) @out(i,j,k,l) = 0;

            for (int i=0; i<poutshape0; i++)
            for (int j=0; j<poutshape1; j++)
            for (int k=0; k<poutshape2; k++)
            for (int l=0; l<poutshape3; l++) {{
                int kx=k*{stride}+{size}/2-{padding};
                int ky=l*{stride}+{size}/2-{padding};
                int bo=1;
                for (int p=kx-{size}/2;p<=kx+{size}/2 && bo;p++)
                for (int q=ky-{size}/2;q<=ky+{size}/2 && bo;q++)
                    if (p>=0 && q>=0 && p<in0shape2 && q<in0shape3)
                    if (@pout(i,j,k,l) == @in0(i,j,p,q)) {{
                        @out(i,j,p,q) += @dout(i,j,k,l);
                        bo=0;
                    }}
            }}
        '''])
    return y

def concat(arr, dim):
    # TODO: low performance when concat lots of vars
    total_dim = 0
    for a in arr:
        total_dim += a.shape[dim]
    cdim = 0
    s = None
    indexes = [ f"i{i}" for i in range(len(a.shape)) ]
    for a in arr:
        shape = list(a.shape)
        shape[dim] = total_dim
        indexes[dim] = f"i{dim}-{cdim}"
        b = a.reindex(shape, indexes)
        # ugly fix for preventing large fused op 
        if len(arr)>=10:
            b.stop_fuse()
        if s is None:
            s = b
        else:
            s += b
        cdim += a.shape[dim]
    return s

def check(bc):
    bc = np.array(bc)
    if ((bc != 1) * (bc != bc.max(0))).sum() > 0:
        raise Exception(f"Shape not match.")
    else:
        return bc.max(0)

def slice_var_index(x, slices):
    if not isinstance(slices, tuple):
        slices = (slices,)
    if isinstance(slices[0], jt.Var):
        if len(slices) == 1 and slices[0].dtype == "bool":
            return (slices[0].where(),)
    bc = []
    ml = -1
    for idx, s in enumerate(slices):
        if isinstance(s, jt.Var):
            shape = s.shape
        elif isinstance(s, np.ndarray):
            shape = list(s.shape)
        elif isinstance(s, list):
            shape = list(np.array(s).shape)
        else:
            continue
        if len(shape) >= ml:
            ml = len(shape)
        bc.append(shape)
    for idx, shape in enumerate(bc):
        if len(shape) < ml:
            shape = (ml - len(shape)) * [1] + shape
            bc[idx] = shape
    if len(bc) >= 1: 
        bc_shape = check(bc)
        ss = []
        for idx, s in enumerate(slices):
            if isinstance(s, np.ndarray) or isinstance(s, list):
                ss.append(jt.array(s).broadcast(bc_shape.tolist()))
            elif isinstance(s, jt.Var):
                ss.append(s.broadcast(bc_shape.tolist()))
            else:
                ss.append(s)
        slices = ss
    out_shape = []
    out_index = []
    shape = x.shape
    cnt_list = 0
    extras_idx = []
    extras = []
    for i in range(len(shape)):
        if i>=len(slices):
            s = slice(None)
        else:
            s = slices[i]
        sp = shape[i]
        j = len(out_shape)
        if isinstance(s, int):
            if s<0: s += sp
            out_index.append(str(s))
        elif isinstance(s, slice):
            if s == slice(None):
                out_shape.append(sp)
                out_index.append(f"i{j}")
                continue
            start = 0 if s.start is None else s.start
            stop = sp if s.stop is None else s.stop
            step = 1 if s.step is None else s.step
            if start<0: start += sp
            if stop<0: stop += sp
            out_shape.append(1+int(max(0, (stop-start-1)//step)))
            out_index.append(f"{start}+i{j}*{step}")
        elif isinstance(s, jt.Var):
            if cnt_list == 0:
                for idx in range(len(bc_shape)):
                    extras_idx.append(f"i{len(out_shape) + idx}")
                out_shape += bc_shape.tolist()
            out_index.append(f"@e{cnt_list}("+ ",".join(extras_idx) + ")")
            cnt_list += 1
            extras.append(s)
        else:
            raise Exception(f"Not support slice {s}")
    if len(out_shape)==0:
        out_shape = [1]
    # Stop fuse both input and output, prevent recompile
    x.stop_fuse()
    return (out_shape, out_index, 0, [], extras)

def slice_var(x, slices):
    reindex_args = slice_var_index(x, slices)
    x.stop_fuse()
    return x.reindex(*reindex_args).stop_fuse()

def setitem(x, slices, value):
    reindex_args = slice_var_index(x, slices)
    reindex_reduce_args = (x.shape, reindex_args[1]) + reindex_args[3:]
    xslice = x.stop_fuse().reindex(*reindex_args).stop_fuse()
    value = jt.broadcast(value, xslice)
    one = jt.broadcast(1, xslice)
    if not isinstance(reindex_args[0][0], jt.Var):
        reindex_args = (x.shape,) + reindex_args[1:]
    mask = one.reindex_reduce("add", *reindex_reduce_args)
    data = value.reindex_reduce("add", *reindex_reduce_args)
    # Stop fuse both input and output, prevent recompile
    out = mask.ternary(data, x).stop_fuse()
    x.assign(out)
    return x

jt.Var.__getitem__ = jt.Var.slice_var = slice_var
jt.Var.__setitem__ = setitem

def adam(model, loss, lr=3e-4, betas=[0.9, 0.999], eps=1e-8):
    ps = jt.find_vars(model)
    gs = jt.grad(loss, ps)
    with jt.var_scope('_'.join([model, 'adam']), unique=True):
        adam_step = jt.make_var([1], init=jt.zeros)
        adam_step += 1
        for p,g in zip(ps,gs):
            m = jt.make_var(p.shape, init=jt.zeros)
            v = jt.make_var(p.shape, init=jt.zeros)
            
            m.assign(betas[0] * m + (1-betas[0]) * g)
            v.assign(betas[1] * v + (1-betas[1]) * g * g)
            step_size = lr * jt.sqrt(1-betas[1]**adam_step) / (1-betas[0] ** adam_step)
            p -= m * step_size / (jt.sqrt(v) + eps)




# ***************************************************************
# Copyright (c) 2020 Jittor. Authors:
#     Guowei Yang <471184555@qq.com>
#     Wenyang Zhou <576825820@qq.com>
#     Meng-Hao Guo <guomenghao1997@gmail.com>
#     Dun Liang <randonlang@gmail.com>.
#
# All Rights Reserved.
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import jittor as jt
from jittor import init, Module
import numpy as np
import math

class Pool(Module):
    def __init__(self, kernel_size, stride=None, padding=0, dilation=None, return_indices=None, ceil_mode=False, op="maximum"):
        assert dilation == None
        assert return_indices == None
        self.kernel_size = kernel_size
        self.op = op
        self.stride = stride if stride else kernel_size
        self.padding = padding
        self.ceil_mode = ceil_mode

    def execute(self, x):
        N,C,H,W = x.shape
        if (self.ceil_mode == False):
            h = (H+self.padding*2-self.kernel_size)//self.stride+1
            w = (W+self.padding*2-self.kernel_size)//self.stride+1
        else:
            h = (H+self.padding*2-self.kernel_size + self.stride - 1)//self.stride+1
            w = (W+self.padding*2-self.kernel_size + self.stride - 1)//self.stride+1

        if (self.op == 'maximum' or self.op == 'minimum'):
            if (self.op == 'maximum'):
                op = 'max'
            else:
                op = 'min'
            out = jt.code([N,C,h,w], x.dtype, [x],
                cuda_src=f'''
                    __global__ static void kernel1(@ARGS_DEF) {{
                        @PRECALC
                        int p3 = threadIdx.x;
                        int s3 = blockDim.x;
                        int p2 = threadIdx.y + blockIdx.x * blockDim.y;
                        int s2 = blockDim.y * gridDim.x;
                        int i1 = blockIdx.y;
                        int i0 = blockIdx.z;
                        for (int i3 = p3; i3 < outshape3; i3 += s3)
                            for (int i2 = p2; i2 < outshape2; i2 += s2) {{
                                int k3 = i3*{self.stride}-{self.padding};
                                int k2 = i2*{self.stride}-{self.padding};
                                int k3_ = min(k3 + {self.kernel_size}, in0shape3);
                                int k2_ = min(k2 + {self.kernel_size}, in0shape2);
                                k3 = max(0, k3);
                                k2 = max(0, k2);
                                @out(i0, i1, i2, i3) = @in0(i0, i1, k2, k3);
                                for (int p = k2; p < k2_; ++p)
                                    for (int q = k3; q < k3_; ++q)
                                        @out(i0, i1, i2, i3) = {op}(@out(i0, i1, i2, i3), @in0(i0, i1, p, q));
                            }}
                    }}
                    int tx = min(1024, outshape3);
                    int ty = min(1024 / tx, outshape2);
                    int bx = (outshape2 - 1) / ty + 1;
                    int by = outshape1;
                    int bz = outshape0;
                    dim3 s1(bx, by, bz);
                    dim3 s2(tx, ty);
                    kernel1<<<s1, s2>>>(@ARGS);
                ''',
                cuda_grad_src=[f'''
                    __global__ static void kernel3(@ARGS_DEF) {{
                        @PRECALC
                        int p3 = threadIdx.x;
                        int s3 = blockDim.x;
                        int p2 = threadIdx.y + blockIdx.x * blockDim.y;
                        int s2 = blockDim.y * gridDim.x;
                        int i1 = blockIdx.y;
                        int i0 = blockIdx.z;
                        for (int i3 = p3; i3 < poutshape3; i3 += s3)
                            for (int i2 = p2; i2 < poutshape2; i2 += s2) {{
                                int k3 = i3*{self.stride}-{self.padding};
                                int k2 = i2*{self.stride}-{self.padding};
                                int k3_ = min(k3 + {self.kernel_size}, in0shape3);
                                int k2_ = min(k2 + {self.kernel_size}, in0shape2);
                                k3 = max(0, k3);
                                k2 = max(0, k2);
                                int bo=1;
                                for (int p = k2; p < k2_ && bo; ++p)
                                    for (int q = k3; q < k3_ && bo; ++q) {{
                                        if (@pout(i0,i1,i2,i3) == @in0(i0,i1,p,q)) {{
                                            atomicAdd(&@out(i0,i1,p,q), @dout(i0,i1,i2,i3));
                                            bo=0;
                                        }}
                                    }}
                            }}
                    }}
                    cudaMemsetAsync(outp, 0, out->size);
                    int tx = min(1024, poutshape3);
                    int ty = min(1024 / tx, poutshape2);
                    int bx = (poutshape2 - 1) / ty + 1;
                    int by = poutshape1;
                    int bz = poutshape0;
                    dim3 s1_(bx, by, bz);
                    dim3 s2_(tx, ty);
                    kernel3<<<s1_, s2_>>>(@ARGS);
                '''],
                cpu_src=f'''
                    for (int i0=0; i0<outshape0; i0++)
                    for (int i1=0; i1<outshape1; i1++)
                    for (int i2=0; i2<outshape2; i2++)
                    for (int i3=0; i3<outshape3; i3++) {{
                        int k2 = i2*{self.stride}-{self.padding};
                        int k3 = i3*{self.stride}-{self.padding};
                        int k2_ = std::min(k2 + {self.kernel_size}, in0shape2);
                        int k3_ = std::min(k3 + {self.kernel_size}, in0shape3);
                        k2 = std::max(0, k2);
                        k3 = std::max(0, k3);
                        @out(i0, i1, i2, i3) = @in0(i0, i1, k2, k3);
                        for (int p = k2; p < k2_; ++p)
                            for (int q = k3; q < k3_; ++q)
                                @out(i0, i1, i2, i3) = std::{op}(@out(i0, i1, i2, i3), @in0(i0, i1, p, q));
                    }}
                ''',
                cpu_grad_src = [f'''
                    for (int i=0; i<outshape0; i++)
                    for (int j=0; j<outshape1; j++)
                    for (int k=0; k<outshape2; k++)
                    for (int l=0; l<outshape3; l++) @out(i,j,k,l) = 0;

                    for (int i0=0; i0<poutshape0; i0++)
                    for (int i1=0; i1<poutshape1; i1++)
                    for (int i2=0; i2<poutshape2; i2++) 
                    for (int i3=0; i3<poutshape3; i3++) {{
                        int k3 = i3*{self.stride}-{self.padding};
                        int k2 = i2*{self.stride}-{self.padding};
                        int k3_ = std::min(k3 + {self.kernel_size}, in0shape3);
                        int k2_ = std::min(k2 + {self.kernel_size}, in0shape2);
                        k3 = std::max(0, k3);
                        k2 = std::max(0, k2);
                        int bo=1;
                        for (int p = k2; p < k2_ && bo; ++p)
                            for (int q = k3; q < k3_ && bo; ++q) {{
                                if (@pout(i0,i1,i2,i3) == @in0(i0,i1,p,q)) {{
                                    @out(i0,i1,p,q) += @dout(i0,i1,i2,i3);
                                    bo=0;
                                }}
                            }}
                    }}
                '''])
            return out
        else:
            xx = x.reindex([N,C,h,w,self.kernel_size,self.kernel_size], [
                "i0", # Nid
                "i1", # Cid
                f"i2*{self.stride}-{self.padding}+i4", # Hid
                f"i3*{self.stride}-{self.padding}+i5", # Wid
            ])
            return xx.reduce(self.op, [4,5])

def pool(x, size, op, padding, stride = 1):
    return Pool(size, stride, padding, op=op)(x)


# ***************************************************************
# Copyright (c) 2020 Jittor. Authors: 
#     Meng-Hao Guo <guomenghao1997@gmail.com>
#     Dun Liang <randonlang@gmail.com>. 
# All Rights Reserved.
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import numpy as np
from urllib import request
import gzip
import pickle
import os
from jittor.dataset.utils import get_random_list, get_order_list, collate_batch
from collections.abc import Sequence, Mapping
import pathlib
from PIL import Image
from jittor_utils.ring_buffer import RingBuffer
import multiprocessing as mp
import signal
from jittor_utils import LOG
import jittor as jt

dataset_root = os.path.join(pathlib.Path.home(), ".cache", "jittor", "dataset")
mp_log_v = os.environ.get("mp_log_v", 0) 
mpi = jt.compile_extern.mpi

class Worker:
    def __init__(self, target, args, buffer_size):
        buffer = mp.Array('c', buffer_size, lock=False)
        self.buffer = RingBuffer(buffer)
        self.p = mp.Process(target=target, args=args+(self.buffer,))
        self.p.daemon = True
        self.p.start()

class Dataset(object):
    '''
    base class for reading data
    
    Example:
        class YourDataset(Dataset):
            def __init__(self):
                super().__init__()
                self.set_attrs(total_len=1024)

            def __getitem__(self, k):
                return k, k*k

        dataset = YourDataset().set_attrs(batch_size=256, shuffle=True)
        for x, y in dataset:
            ......
    '''
    def __init__(self,
                 batch_size = 16,
                 shuffle = False,
                 drop_last = False,
                 num_workers = 0,
                 buffer_size = 512*1024*1024):
        super().__init__()
        self.total_len = None
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.drop_last = drop_last
        self.num_workers = num_workers
        self.buffer_size = buffer_size

    def __getitem__(self, index):
        raise NotImplementedError

    def __len__(self):
        assert self.total_len >= 0
        assert self.batch_size > 0
        real_len = (self.total_len-1)//mpi.world_size()+1 if mpi \
            else self.total_len
        if self.drop_last:
            return real_len // self.batch_size
        return (real_len-1) // self.batch_size + 1

    def set_attrs(self, **kw):
        '''set attributes of dataset, equivalent to setattr
        
        Attrs:
            batch_size(int): batch size, default 16.
            totol_len(int): totol lenght.
            shuffle(bool): shuffle at each epoch, default False.
            drop_last(bool): if true, the last batch of dataset
                might smaller than batch_size, default True.
            num_workers: number of workers for loading data
            buffer_size: buffer size for each worker in bytes,
                default(512MB).
        '''
        for k,v in kw.items():
            assert hasattr(self, k), k
            setattr(self, k, v)
        return self

    def to_jittor(self, batch):
        if isinstance(batch, np.ndarray):
            return jt.array(batch)
        assert isinstance(batch, Sequence)
        new_batch = []
        for a in batch:
            if isinstance(a, np.ndarray) or \
                isinstance(a, int) or \
                isinstance(a, float):
                new_batch.append(jt.array(a))
            else:
                new_batch.append(a)
        return new_batch

    def collate_batch(self, batch):
        return collate_batch(batch)

    def terminate(self):
        if hasattr(self, "workers"):
            for w in self.workers:
                w.p.terminate()
    
    def _worker_main(self, worker_id, buffer):
        try:
            gid_obj = self.gid.get_obj()
            gid_lock = self.gid.get_lock()
            while True:
                with gid_lock:
                    while gid_obj.value >= self.batch_len:
                        self.num_idle.value += 1
                        self.num_idle_c.notify()
                        self.gidc.wait()
                        self.num_idle.value -= 1
                    cid = gid_obj.value
                    self.idmap[cid] = worker_id
                    gid_obj.value += 1
                    self.gidc.notify()
                batch = []
                if mp_log_v:
                    print(f"#{worker_id} {os.getpid()} load batch", cid*self.batch_size, min(self.real_len, (cid+1)*self.batch_size))
                for i in range(cid*self.batch_size, min(self.real_len, (cid+1)*self.batch_size)):
                    batch.append(self[self.index_list[i]])
                batch = self.collate_batch(batch)
                if mp_log_v:
                    print(f"#{worker_id} {os.getpid()} send", type(batch).__name__, [ type(b).__name__ for b in batch ], buffer)
                buffer.send(batch)
        except:
            os.kill(os.getppid(), signal.SIGINT)
            raise

    def _stop_all_workers(self):
        # wait until all workers idle
        if self.num_idle.value < self.num_workers:
            with self.gid.get_lock():
                self.gid.get_obj().value = self.batch_len
                if mp_log_v:
                    print("idle num", self.num_idle.value)
                while self.num_idle.value < self.num_workers:
                    self.num_idle_c.wait()
                    if mp_log_v:
                        print("idle num", self.num_idle.value)
        # clean workers' buffer
        for w in self.workers:
            w.buffer.clear()
            
    def _init_workers(self):
        self.index_list = mp.Array('i', self.real_len, lock=False)
        workers = []
        # batch id to worker id
        self.idmap = mp.Array('i', self.batch_len, lock=False)
        # global token index
        self.gid = mp.Value('i', self.batch_len)
        # global token index condition
        self.gidc = mp.Condition(self.gid.get_lock())
        # number of idle workers
        self.num_idle = mp.Value('i', 0, lock=False)
        # number of idle workers condition
        self.num_idle_c = mp.Condition(self.gid.get_lock())
        for i in range(self.num_workers):
            w = Worker(target=self._worker_main, args=(i,), 
                       buffer_size=self.buffer_size)
            workers.append(w)
        self.workers = workers
        self.index_list_numpy = np.ndarray(dtype='int32', shape=self.real_len, buffer=self.index_list)

    def __del__(self):
        if mp_log_v:
            print("dataset deleted")
        self.terminate()

    def __iter__(self):
        if self.shuffle == False:
            index_list = get_order_list(self.total_len)
        else:
            index_list = get_random_list(self.total_len)
        
        # scatter index_list for all mpi process
        # scatter rule:
        # [........]
        #  000111
        #       222
        # make sure each process has the same len
        if mpi:
            index_list = np.int32(index_list)
            mpi.broadcast(index_list, 0)
            real_len = (self.total_len - 1) // mpi.world_size() + 1
            offset = mpi.world_rank() * real_len
            if offset + real_len > self.total_len:
                offset -= offset + real_len - self.total_len
            index_list = index_list[offset:offset+real_len]
            self.real_len = real_len
            assert real_len == len(index_list)
        else:
            self.real_len = self.total_len
            
        self.batch_len = len(self)
        if "batch_len" in os.environ:
            self.batch_len = int(os.environ["batch_len"])
        
        if not hasattr(self, "workers") and self.num_workers:
            self._init_workers()
        
        if self.num_workers:
            self._stop_all_workers()
            self.index_list_numpy[:] = index_list
            gid_obj = self.gid.get_obj()
            gid_lock = self.gid.get_lock()
            with gid_lock:
                gid_obj.value = 0
                self.gidc.notify_all()
            for i in range(self.batch_len):
                # try not get lock first
                if gid_obj.value <= i:
                    with gid_lock:
                        if gid_obj.value <= i:
                            if mp_log_v:
                                print("wait")
                            self.gidc.wait()
                worker_id = self.idmap[i]
                w = self.workers[worker_id]
                if mp_log_v:
                    print(f"#{worker_id} {os.getpid()} recv buffer", w.buffer)
                batch = w.buffer.recv()
                if mp_log_v:
                    print(f"#{worker_id} {os.getpid()} recv", type(batch).__name__, [ type(b).__name__ for b in batch ])
                batch = self.to_jittor(batch)
                yield batch
        else:
            batch_data = []
            for idx in index_list:
                batch_data.append(self[int(idx)])
                if len(batch_data) == self.batch_size:
                    batch_data = self.collate_batch(batch_data)
                    batch_data = self.to_jittor(batch_data)
                    yield batch_data
                    batch_data = []

            # depend on drop_last
            if not self.drop_last and len(batch_data) > 0:
                batch_data = self.collate_batch(batch_data)
                batch_data = self.to_jittor(batch_data)
                yield batch_data


class ImageFolder(Dataset):
    """A image classify dataset, load image and label from directory:
    
        root/label1/img1.png
        root/label1/img2.png
        ...
        root/label2/img1.png
        root/label2/img2.png
        ...
    Args:
        root(string): Root directory path.

     Attributes:
        classes(list): List of the class names.
        class_to_idx(dict): map from class_name to class_index.
        imgs(list): List of (image_path, class_index) tuples
    """
    def __init__(self, root, transform=None):
        # import ipdb; ipdb.set_trace()
        super().__init__()
        self.root = root
        self.transform = transform
        self.classes = sorted([d.name for d in os.scandir(root) if d.is_dir()])
        self.class_to_idx = {v:k for k,v in enumerate(self.classes)}
        self.imgs = []
        image_exts = set(('.jpg', '.jpeg', '.png', '.bmp', '.tif', '.tiff'))
        
        for i, class_name in enumerate(self.classes):
            class_dir = os.path.join(root, class_name)
            for dname, _, fnames in sorted(os.walk(class_dir, followlinks=True)):
                for fname in sorted(fnames):
                    if os.path.splitext(fname)[-1].lower() in image_exts:
                        path = os.path.join(class_dir, fname)
                        self.imgs.append((path, i))
        LOG.i(f"Found {len(self.classes)} classes and {len(self.imgs)} images.")
        self.set_attrs(total_len=len(self.imgs))
        
    def __getitem__(self, k):
        with open(self.imgs[k][0], 'rb') as f:
            img = Image.open(f).convert('RGB')
            if self.transform:
                img = self.transform(img)
            return img, self.imgs[k][1]



# ***************************************************************
# Copyright (c) 2020 Jittor. Authors: 
#     Guoye Yang <498731903@qq.com>
#     Dun Liang <randonlang@gmail.com>. 
# All Rights Reserved.
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import os
os.environ['OPENBLAS_NUM_THREADS'] = '1'

import unittest
import time
import jittor as jt
from jittor import LOG
import math
import numpy as np
from .test_core import expect_error
from .test_fused_op import retry

def check_cache_code(fname):
    check_code = True
    error_line_num = -1
    with open(fname) as f:
        lines = f.readlines()
        for i in range(len(lines)):
            if ('memory_checker->check_hit(' in lines[i]):
                continue
            code = lines[i]
            address_pos = []
            for j in range(len(code)):
                if code[j] == '[':
                    address_pos.append(j)
                if code[j] == ']':
                    sp = address_pos[-1] - 1
                    address_pos = address_pos[:-1]
                    if sp>=4 and code[sp-4:sp+1]=="shape":
                        continue
                    while ((sp >= 0) and ((code[sp] >= 'A' and code[sp] <= 'Z') or (code[sp] >= 'a' and code[sp] <= 'z') or 
                    (code[sp] >= '0' and code[sp] <= '9') or code[sp] == '_' or code[sp] == '.' or (sp > 0 and code[sp] == '>' and code[sp - 1] == '-'))):
                        if (sp > 0 and code[sp] == '>' and code[sp - 1] == '-'):
                            sp -= 2
                        else:
                            sp -= 1
                    sp += 1
                    check_var = code[sp:j + 1]
                    temp_i = i - 1
                    have_check = False
                    while (temp_i >= 0 and 'memory_checker->check_hit(' in lines[temp_i]):
                        if check_var in lines[temp_i]:
                            have_check = True
                            break
                        temp_i -= 1
                    if (not have_check):
                        check_code = False
                        error_line_num = i
                        break
            if (not check_code):
                break
    assert check_code, "check cache not found in line " + str(error_line_num) + " of file " + fname

class TestCache(unittest.TestCase):   
    def test_reduce(self):
        @retry(10)
        def check(n, m, reduce_dim, cache_report_, error_rate_threshold):
            a = jt.random([n,m])
            a.sync()
            with jt.profile_scope(compile_options = {
                "check_cache": 1, "replace_strategy": 1, "page_size": 4 << 10, #2 << 20
                "vtop": 0,
                "tlb_size": 64, "tlb_ways": 4, "tlb_line_size": 1,
                "L1_size": 32 << 10, "L1_ways": 8, "L1_line_size": 64,
                "L2_size": 256 << 10, "L2_ways": 8, "L2_line_size": 64,
                "L3_size": 15 << 20, "L3_ways": 20, "L3_line_size": 64
            }, enable_tuner=0) as report:
                c = a.sum(reduce_dim)
                c.sync()
            
            check_cache_code(report[1][1])
            cache_report = report[-1][-5:]
            for i in range(len(cache_report)):
                cache_report[i] = int(cache_report[i])
            for i in range(len(cache_report)):
                assert abs(cache_report[i] - cache_report_[i]) <= int(cache_report_[i] * error_rate_threshold), "cache report error: " + report[-2][-(len(cache_report) - i)] + " error, " + str(cache_report[i]) + "!=" + str(cache_report_[i])
        error_threshold = 0.02
        check(100, 10000, 0, [3010004, 989, 125729, 63129, 63129], error_threshold)
        check(100, 10000, 1, [3000104, 981, 62510, 62510, 62510], error_threshold)
        check(10, 98765, 0, [3061719, 2034, 129645, 129645, 67905], error_threshold)
        check(10, 98765, 1, [2962964, 969, 61733, 61733, 61733], error_threshold)
        check(7779, 97, 0, [2263790, 740, 47170, 47170, 47170], error_threshold)
        check(7779, 97, 1, [2271472, 748, 47650, 47650, 47650], error_threshold)
        check(1024, 1024, 0, [3146756, 1029, 65603, 65603, 65603], error_threshold)
        check(1024, 1024, 1, [3146756, 1028, 65603, 65603, 65603], error_threshold)
    
if __name__ == "__main__":
    unittest.main()



# ***************************************************************
# Copyright (c) 2020 Jittor. Authors: Dun Liang <randonlang@gmail.com>. All Rights Reserved.
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import time
from .test_core import expect_error
import os

mid = 0
if os.uname()[1] == "jittor-ce":
    mid = 1

class TestNanoString(unittest.TestCase):
    def test(self):
        dtype = jt.NanoString
        t = time.time()
        n = 1000000
        for i in range(n):
            dtype("float")
        t = (time.time() - t)/n
        # t is about 0.01 for 100w loop
        # 92ns one loop
        assert t < [1.5e-7, 1.7e-7][mid], t

        assert (jt.hash("asdasd") == 4152566416)
        assert str(jt.NanoString("float"))=="float"
        # pybind11: 7
        # Tuple call: 1.3
        # fast call (with or with not): 0.9
        # init call 1.5
        # int init: 1.2
        # dtype init(cache): 0.75
        # final: 1.0

if __name__ == "__main__":
    unittest.main()